#include <stdio.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include <stdlib.h>

#define NL(flnm,dumch) {do dumch=fgetc(flnm); while (dumch!='\n' && dumch!=EOF);}
#define ntmapstep(x,y,a,b) {y-=(b)*sinl(pi2*(x)); x+=(a)-(a)*(y)*(y);}
#define ntmapdm(dd,x,y,a,b,tmp1,tmp2) {tmp1=-2.0L*(a)*((y)-(b)*sinl(pi2*(x))); tmp2=-pi2*(b)*cosl(pi2*(x)); dd[0][0]=1.0L+tmp1*tmp2; dd[0][1]=tmp1; dd[1][0]=tmp2; dd[1][1]=1.0L;}
#define smoothen(x,y) {if ((x)>0.0L) y=((x)/(1.0L+(x))); else y=((x)/(1.0L-(x)));}
#define SIGN(a,b) ((b) >= 0.0 ? fabsl(a) : -fabsl(a))
#define FMAX(a,b) ((a) > (b) ? (a) : (b))
#define IPOW3(a) ((int)(pow(3.0L, (long double)(a))+0.5L)) 
#define MOD(a) (dreml(a,1.0L)>0.5L ? dreml(a,1.0L)-1.0L: dreml(a,1.0L)) 
#define ERROR 123.456789L
#define UNDEFINED -123456.789L

const long double SMALL=1.0e-10L;
const int SUPPERROR=15;

void makeinitials(void);
void prepoutfiles(void);
void maincalc(void);
void selectfiterfuncs(int, int);
long double rootybfunc(long double, int, unsigned long int, unsigned long int, long double, long double, 
		       long double, long double);
long double residue(long double, long double, long double, int, int, long double *, long double *, 
		    long double *, long double *, int *);
long double findx0(long double, long double, int);
long double fiter2funcoe0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoo0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funceo0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoe1(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoo1(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funceo1(long double, long double, int, unsigned long int, unsigned long int);
void zbrak2(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double, int, unsigned long int, unsigned long int, 
	   long double, long double, int, long double [], long double [], int *);
int zbrac2(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double, int, unsigned long int, unsigned long int, long double, long double,
	   long double *, long double *);
long double zbrent2(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
		   long double, int, unsigned long int, unsigned long int, long double, long double);
void mkstr(char *, unsigned long int);

long double pi2;
long double omega[50], a0, a00[50], b0, b00[50], ymin[5], ymax[5], yguess0[5][50], yguess[5], ygiters;
int sln, slnc, slnmin, slnmax, qpi, qp0, qpx, ab_no, ybraks, ydepth;
unsigned long int qq[50], pp[50], optcount=0;
char iterfilename[160], iterfilenamebase[80];

FILE *fl1, *fl;
long double (*fiterfunc)(long double, long double, int, unsigned long int, unsigned long int);

int main(int argc, char *argv[])
{
  int i;   

  /* read input */
  if ( (fl=fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "*** error opering in-file  %s  ***\n", argv[1]);
    exit(1);
  }
  makeinitials();
  fclose(fl);

  /* prepare output files */
  prepoutfiles();

  for (i=0; i<ab_no; i++) {
      a0=a00[i];
      b0=b00[i];
      for (sln=slnmin;sln<=slnmax;sln++) {
	  yguess[sln]=yguess0[sln][i];
	  if (yguess[i]>ymax[i] || yguess[i]<ymin[i]) yguess[i]=(ymin[i]+ymax[i])/2.0L;
      }
      maincalc();
  }
  
  fclose(fl1);
  exit(0); 
} 


void makeinitials(void)      /* read and adjust initial values */
{
  FILE *fl;
  char dummy;
  int i, cfn, cfp, cfa[50];
  unsigned long int j;
  long double tmp, tmpx, tmpy, tmpc1, tmpc2;
  
  pi2=8.0L*atanl(1.0L);

  /* read initials */
  fscanf(fl,"%d", &sln); NL(fl, dummy);                /* symmetry line number */
  fscanf(fl,"%Lf %Lf", &ymin[1], &ymax[1]);  NL(fl, dummy);  /* s1 y-range & ycrit */
  fscanf(fl,"%Lf %Lf", &ymin[2], &ymax[2]);  NL(fl, dummy);  /* s2 y-range & ycrit */
  fscanf(fl,"%Lf %Lf", &ymin[3], &ymax[3]);  NL(fl, dummy);  /* s3 y-range & ycrit */
  fscanf(fl,"%Lf %Lf", &ymin[4], &ymax[4]);  NL(fl, dummy);  /* s4 y-range & ycrit */
  fscanf(fl,"%Le", &ygiters); NL(fl, dummy);           /* iterations for yguess on other slns*/
  fscanf(fl,"%d %d", &ybraks, &ydepth);  NL(fl, dummy);/* # of braks if initial y-bracket fails, # to check */
  fscanf(fl,"%s", iterfilenamebase); NL(fl, dummy);    /* outfilename */
  fscanf(fl,"%d %d", &qp0, &qpx);  NL(fl, dummy);      /* start and end number of p/q values in FLS*/
  fscanf(fl,"%d %d", &cfn, &cfp); NL(fl, dummy);       /* number of cf coeff. and period */
  for (i=0;i<cfn;i++) fscanf(fl,"%d ", &cfa[i]); NL(fl, dummy); 
  fscanf(fl,"%d %d", &ab_no, &slnc);  NL(fl, dummy);   /* # (a,b, y_si) values below, i */
  for (i=0;i<ab_no;i++) {
      fscanf(fl,"%Lf %Lf %Lf", &a00[i], &b00[i], &yguess0[1][i]); NL(fl, dummy); 
      yguess0[2][i]=yguess0[1][i]; yguess0[3][i]=yguess0[1][i]; yguess0[4][i]=yguess0[1][i];
  }
  
  for (i=1;i<=4;i++) {
      if (ymin[i]>ymax[i]) { tmp=ymin[i]; ymin[i]=ymax[i];ymax[i]=tmp; }      
  }
      
  if (sln==0) {slnmin=1; slnmax=4;}
  else {slnmin=sln; slnmax=sln;}
  if (qp0<1) qp0=1;
  if (qpx<qp0) qpx=qp0;
  if (cfp>cfn) cfp=1;

  for (i=cfn;i<=qpx;i++) cfa[i]=cfa[cfn-cfp+(i-cfn)%cfp];
  qq[0]=1; pp[0]=cfa[0];
  omega[0]=((long double)(qq[0]))/((long double)(pp[0]));
  qq[1]=cfa[1]; pp[1]=pp[0]*cfa[1]+1;
  omega[1]=((long double)(qq[1]))/((long double)(pp[1]));
  for (i=2;i<=qpx;i++) {
      qq[i]=cfa[i]*qq[i-1]+qq[i-2];
      pp[i]=cfa[i]*pp[i-1]+pp[i-2];      
      omega[i]=((long double)(qq[i]))/((long double)(pp[i]));
  }
  for (i=0;i<ab_no;i++) {
      for (sln=1;sln<=4;sln++) {
	  switch (sln) {
	      case 1: {tmpc1=0.0L; tmpc2=0.0L; break;}
	      case 2: {tmpc1=0.5L; tmpc2=0.0L; break;}
	      case 3: {tmpc1=a00[i]/2.0L; tmpc2=a00[i]/2.0L; break;}
	      case 4: {tmpc1=a00[i]/2.0L+0.5L; tmpc2=a00[i]/2.0L; break;}
	      default: {tmpc1=0.0L; tmpc2=0.0L;}
	  }	      
	  if (sln!=slnc) {
	      tmp=1.0L; tmpy=yguess0[slnc][i];
	      switch (slnc) {
		  case 1: {tmpx=0.0L; break;}
		  case 2: {tmpx=0.5L; break;}
		  case 3: {tmpx=a00[i]/2.0L*(1.0L-tmpy*tmpy); break;}
		  case 4: {tmpx=a00[i]/2.0L*(1.0L-tmpy*tmpy)+0.5L; break;}
		  default: tmpx=0.0L;
	      }	      
	      for (j=ygiters;j>0;j--) {
		  ntmapstep(tmpx, tmpy, a00[i], b00[i]);
		  tmpx=dreml(tmpx,1.0L);
		  if (fabsl(dreml(tmpx-tmpc1+tmpc2*tmpy*tmpy,1.0L))<tmp) {
		      tmp=fabsl(dreml(tmpx-tmpc1+tmpc2*tmpy*tmpy,1.0L));
		      yguess0[sln][i]=tmpy;
		  }
	      }      
	  }
      }
      
  }
  

}


void prepoutfiles(void)
{
  char ctmp[60];
  int i;
    
  strcpy(iterfilename,iterfilenamebase);
  strcpy(ctmp,""); mkstr(&ctmp[0],qp0); strcat(iterfilename,ctmp);
  strcat(iterfilename,"_");
  strcpy(ctmp,""); mkstr(&ctmp[0],qpx); strcat(iterfilename,ctmp);

  fl1=fopen(iterfilename, "w");
  fprintf(fl1,"# Symm.l.: %2d \n", sln);
  fprintf(fl1,"# ymin1:   %10.7Lf , ymax1:   %10.7Lf \n", ymin[1], ymax[1]); 
  fprintf(fl1,"# ymin1:   %10.7Lf , ymax1:   %10.7Lf \n", ymin[2], ymax[2]); 
  fprintf(fl1,"# ymin1:   %10.7Lf , ymax1:   %10.7Lf \n", ymin[3], ymax[3]); 
  fprintf(fl1,"# ymin1:   %10.7Lf , ymax1:   %10.7Lf \n", ymin[4], ymax[4]); 
  fprintf(fl1,"# ygiters: %10.3Le\n", ygiters); 
  fprintf(fl1,"# ybraks:  %10d , ydepth:  %10d\n", ybraks, ydepth);
  fprintf(fl1,"# "); 
  for (i=qp0; i<=qpx; i++) fprintf(fl1,"[%d]=%lu/%lu, ", i, qq[i], pp[i]); 
  fprintf(fl1,"~%.18Lf\n", ((long double)(qq[qpx]))/((long double)(pp[qpx]))); 
  for (i=0; i<ab_no; i++)  
      fprintf(fl1,"# a0: %23.18Lf, b0: %23.18Lf, y1: %23.18Lf, y2: %23.18Lf, y3: %23.18Lf, y4: %23.18Lf\n", 
	      a00[i], b00[i], yguess0[1][i], yguess0[2][i], yguess0[3][i], yguess0[4][i]); 
  fprintf(fl1,"#%4s%3s%23s%23s%23s%23s%23s%8s%8s%8s\n", 
	  "[n]", "sln", "Res", "a", "b", "x", "y", "lg(dR)", "lg(dx)", "lg(dy)");
}

void maincalc() 
{
    long double upsol[5][50]={{0},{0}}, downsol[5][50]={{0},{0}};
    long double ydownmin[5]={0}, ydownmax[5]={0}, yupmin[5]={0}, yupmax[5]={0};
    int i, ok, brakmax=10, midsln;
    long double braku[11]={0}, brakl[11]={0};
    long double dresid[5]={0}, dxerr[5]={0}, dyerr[5]={0}, dxmid, dymid;
    long double uresid[5]={0}, uxerr[5]={0}, uyerr[5]={0}, uxmid, uymid;
    long double dreserr[5]={0}, ureserr[5]={0}, dummy1, dummy2, dummy3, dummy4;

    for (sln=slnmin;sln<=slnmax;sln++) {
	ydownmin[sln]=ymin[sln]; ydownmax[sln]=yguess[sln]; 
	yupmin[sln]=yguess[sln]; yupmax[sln]=ymax[sln]; 
    }
    
    for (qpi=qp0; qpi<=qpx;qpi+=2){
	midsln=0;
	for (sln=slnmin;sln<=(slnmax>slnmin ? slnmax/2 : slnmax) ;sln++) {
	    if (sln==midsln) sln++;
	    selectfiterfuncs(qpi,sln);
	    
	    ok=1;
	    downsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], 
					 ydownmin[sln], (ydownmin[sln]+ydownmax[sln])/2.0L, ydownmax[sln], 
					 (ydownmax[sln]-ydownmin[sln])/((long double) (ybraks)));
	    upsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], 
				       yupmin[sln], (yupmin[sln]+yupmax[sln])/2.0L, yupmax[sln], 
				       (yupmax[sln]-yupmin[sln])/((long double) (ybraks)));
	    
	    if (fabsl(downsol[sln][qpi]-ERROR)<SMALL || fabsl(upsol[sln][qpi]-ERROR)<SMALL) {
		optcount++; brakmax=ydepth;
		zbrak2(*fiterfunc, b0, sln, qq[qpi], pp[qpi], 
		       ydownmin[sln], yupmax[sln], ybraks, &brakl[0], &braku[0], &brakmax);
		if (brakmax==2) {
		    downsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], 
						 brakl[1], (brakl[1]+braku[1])/2.0L, braku[1], 
						 (braku[1]-brakl[1])/((long double) (ybraks)));
		    upsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], 
					       brakl[2], (brakl[2]+braku[2])/2.0L, braku[2], 
					       (braku[2]-brakl[2])/((long double) (ybraks)));
		    if (fabsl(downsol[sln][qpi]-ERROR)<SMALL || fabsl(upsol[sln][qpi]-ERROR)<SMALL) {
			ok=0;
			printf("(%d) Huh? Right number of brackets, but no solutions?!?!\n", sln);
		    } else ok=1;
		} else if (brakmax<2) {
		    ok=0;
		    printf("(%d) Not enough brackets: wrong intervals or too small ybraks?\n", sln);
		} else {
		    printf("(%d) Warning: Too many brackets: too chaotic?\n", sln);
		    for (i=1; i<brakmax; i++) {
			ydownmin[sln]=brakl[1]; ydownmax[sln]=braku[1];
			yupmin[sln]=brakl[2]; yupmax[sln]=braku[2];
			if (brakl[i]<yguess[i]) {
			    ydownmin[sln]=brakl[i]; ydownmax[sln]=braku[i];
			    yupmin[sln]=brakl[i+1]; yupmax[sln]=braku[i+1];
			}
		    }
		    
		    downsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], ydownmin[sln], 
						 (ydownmin[sln]+ydownmax[sln])/2.0L, ydownmax[sln], 
						 (ydownmax[sln]-ydownmin[sln])/((long double) (ybraks)));
		    upsol[sln][qpi]=rootybfunc(b0, sln, qq[qpi], pp[qpi], yupmin[sln], 
					       (yupmin[sln]+yupmax[sln])/2.0L, yupmax[sln], 
					       (yupmax[sln]-yupmin[sln])/((long double) (ybraks)));
		    
		    if (fabsl(downsol[sln][qpi]-ERROR)<SMALL || fabsl(upsol[sln][qpi]-ERROR)<SMALL) {
			ok=0;
			printf("(%d) Huh? Brackets found, but no solutions?!?!\n", sln);
		    } else ok=1;
		    
		}
	    }
	    
	    if (ok<=0) break;
	    dresid[sln]=residue(a0, b0, downsol[sln][qpi], sln, qpi, 
				&dxerr[sln], &dyerr[sln], &dxmid, &dymid, &midsln);
	    uresid[sln]=residue(a0, b0, upsol[sln][qpi], sln, qpi, 
				&uxerr[sln], &uyerr[sln], &uxmid, &uymid, &midsln);

	    dreserr[0]=residue(a0, b0, downsol[sln][qpi]-dyerr[sln], sln, qpi, 
			       &dummy1, &dummy2, &dummy3, &dummy4, &midsln);
	    dreserr[sln]=dresid[sln]-dreserr[0];
	    dreserr[0]=residue(a0, b0, downsol[sln][qpi]+dyerr[sln], sln, qpi, 
			       &dummy1, &dummy2, &dummy3, &dummy4, &midsln);
	    dreserr[sln]=(fabsl(dresid[sln]-dreserr[0])>fabsl(dreserr[sln]) ?
			  dresid[sln]-dreserr[0] : dreserr[sln]);

	    ureserr[0]=residue(a0, b0, upsol[sln][qpi]+uyerr[sln], sln, qpi, 
			       &dummy1, &dummy2, &dummy3, &dummy4, &midsln);
	    ureserr[sln]=uresid[sln]-ureserr[0];
	    ureserr[0]=residue(a0, b0, upsol[sln][qpi]-uyerr[sln], sln, qpi, 
			       &dummy1, &dummy2, &dummy3, &dummy4, &midsln);
	    ureserr[sln]=(fabsl(uresid[sln]-ureserr[0])>fabsl(ureserr[sln]) ?
			  uresid[sln]-ureserr[0] : ureserr[sln]);
	    
	    if (dymid<uymid){
		downsol[midsln][qpi]=dymid;
		upsol[midsln][qpi]=uymid;
		dyerr[midsln]=dyerr[sln]; dxerr[midsln]=dxerr[sln];
		uyerr[midsln]=uyerr[sln]; uxerr[midsln]=uxerr[sln];
		dresid[midsln]=dresid[sln]; dreserr[midsln]=dreserr[sln];
		uresid[midsln]=uresid[sln]; ureserr[midsln]=ureserr[sln];
	    } else {
		downsol[midsln][qpi]=uymid;
		upsol[midsln][qpi]=dymid;
		dyerr[midsln]=uyerr[sln]; dxerr[midsln]=uxerr[sln];
		uyerr[midsln]=dyerr[sln]; uxerr[midsln]=dxerr[sln];
		dresid[midsln]=uresid[sln]; dreserr[midsln]=ureserr[sln];
		uresid[midsln]=dresid[sln]; ureserr[midsln]=dreserr[sln];
	    }
	}

	if (ok<=0) break;    
	for (sln=slnmin;sln<=slnmax;sln++) {
	    ydownmin[sln]=downsol[sln][qpi]; yupmax[sln]=upsol[sln][qpi];
	    
	    if (yguess[sln]>yupmax[sln] || yguess[sln]<ydownmin[sln]) {
		if (qpi-qp0>=12){
		    dummy1=(yguess[sln]-upsol[sln][qpi-12])/(downsol[sln][qpi-12]-yguess[sln]);
		    yguess[sln]=(dummy1*ydownmin[sln]+yupmax[sln])/(1.0L+dummy1);
		} else yguess[sln]=(ydownmin[sln]+yupmax[sln])/2.0L; 
		if (yguess[sln]>yupmax[sln] || yguess[sln]<ydownmin[sln])
		    yguess[sln]=(ydownmin[sln]+yupmax[sln])/2.0L; 
		ydownmax[sln]=yguess[sln]; yupmin[sln]=yguess[sln];
	    }
	    	    
	    fprintf(fl1,"%5d%3d%23.18Lf%23.18Lf%23.18Lf%23.18Lf%23.18Lf%8.3Lf%8.3Lf%8.3Lf\n", 
		    qpi, -sln, dresid[sln], a0, b0, 
		    MOD(findx0(a0, downsol[sln][qpi], sln)), downsol[sln][qpi], 
		    log10l(fabsl(dreserr[sln])), log10l(fabsl(dxerr[sln])), log10l(fabsl(dyerr[sln])));
	    fprintf(fl1,"%5d%3d%23.18Lf%23.18Lf%23.18Lf%23.18Lf%23.18Lf%8.3Lf%8.3Lf%8.3Lf%5.2Lf\n", 
		    qpi, sln, uresid[sln], a0, b0, 
		    MOD(findx0(a0, upsol[sln][qpi], sln)), upsol[sln][qpi], 
		    log10l(fabsl(ureserr[sln])), log10l(fabsl(uxerr[sln])), log10l(fabsl(uyerr[sln])),
		    (yguess[sln]-upsol[sln][qpi])/(downsol[sln][qpi]-yguess[sln]));
	    fflush(fl1);
	}
    }
    fprintf(fl1,"\n");
    printf("%u\n", optcount);

}

long double rootybfunc(long double bbb, int sss, unsigned long int qqq, unsigned long int ppp,
		       long double y0tmp, long double yyy, long double yxxtmp, long double ybracsize0)
{
    int brakno=2, ok=0;
    long double bracl=yyy, bracu=yyy+ybracsize0;

    if ((*fiterfunc)(bbb, y0tmp, sss, qqq, ppp)*(*fiterfunc)(bbb, yxxtmp, sss, qqq, ppp)>= 0.0)
	ok=zbrac2(*fiterfunc, bbb, sss, qqq, ppp, y0tmp, yxxtmp, &bracl, &bracu);
    else { bracl=y0tmp; bracu=yxxtmp; ok=1;}
    if (ok>0) return zbrent2(*fiterfunc, bbb, sss, qqq, ppp, bracl, bracu);
    else return ERROR;
}

/* find residues of a known periodic orbit, with opt. output to fl2 and *tmp on I0 */
long double residue(long double aaa, long double bbb, long double yy0, int symmln, int ii, 
		    long double *xerr, long double *yerr, 
		    long double *xmid, long double *ymid, int *smid)
{
  long double xxx, yyy, dpq[2][2], lm[2][2], lm0[2][2], tmp1, tmp2;
  unsigned long int i, j, k, halfiters;

  if (2*(pp[ii]/2)==pp[ii]) {            /* p even (-> q odd) */
      halfiters=pp[ii]/2;
      *smid=(symmln<=2? (symmln==1 ? 2:1) : (symmln==3 ? 4:3));   
  }
  else {                               /* p odd  */
      if (2*(qq[ii]/2)==qq[ii])  {     /* q even */
	  if (symmln<=2) {
	      halfiters=(pp[ii]+1)/2;
	      *smid=symmln+2;
	  } else {
	      halfiters=(pp[ii]-1)/2;
	      *smid=symmln-2;
	  }	  
      } else {                         /* q odd  */
	  if (symmln<=2) {
	      halfiters=(pp[ii]+1)/2;
	      *smid=5-symmln;
	  } else {
	      halfiters=(pp[ii]-1)/2;
	      *smid=5-symmln;
	  }	  
      }
  }

  xxx=findx0(aaa, yy0, symmln);
  yyy=yy0;  

  ntmapdm(lm, xxx, yyy, aaa ,bbb, tmp1, tmp2);

  for (k=1;k<pp[ii];k++) {
    ntmapstep(xxx, yyy, aaa, bbb); xxx=dreml(xxx,1.0L);
    if (k==halfiters) {*xmid=xxx; *ymid=yyy;}
    ntmapdm(dpq, xxx, yyy, aaa ,bbb, tmp1, tmp2);
    for (i=0;i<=1;i++) for (j=0;j<=1;j++) lm0[i][j]=lm[i][j]; 
    for (i=0;i<=1;i++) for (j=0;j<=1;j++) lm[i][j]=dpq[i][0]*lm0[0][j]+dpq[i][1]*lm0[1][j];
  }

  ntmapstep(xxx, yyy, aaa, bbb); xxx=dreml(xxx,1.0L);
  *xerr=xxx-dreml(findx0(aaa, yy0, symmln),1.0L);
  if (fabsl(*xerr)>0.5L) *xerr=dreml(xxx-0.5L,1.0L)-dreml(findx0(aaa, yy0, symmln)-0.5L,1.0L);
  *yerr=yyy-yy0;
  return (2.0L-lm[0][0]-lm[1][1])/4.0L;
}

/* find x0(y0) so that point lies on symmetry line */
long double findx0(long double aa0, long double yy0, int symmln)
{
  long double tmp;

  switch (symmln) {
    case 1: {tmp=0.0L; break;}
    case 2: {tmp=0.5L; break;}
    case 3: {tmp=aa0/2.0L*(1.0L-yy0*yy0); break;}
    case 4: {tmp=aa0/2.0L*(1.0L-yy0*yy0)+0.5L; break;}
    default: {tmp=0.0L; printf("Uhoh - sln...\n");}
  }

  /* x modulo 1
  while (tmp> 0.5L-SMALL) tmp=tmp-1.0L;
  while (tmp< -0.5-SMALL) tmp=tmp+1.0L; */

  return tmp;
}

long double fiter2funcoe0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==1) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y; 
  for (i=1;i<=ppp/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-0.5L,xxtmp);}
  else {smoothen(xxtmp,xxtmp);}
  return xxtmp;
}

long double fiter2funcoe1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, j, jj, qf, pf;
  register long double xxtmp, yytmp;

  if (sss==3) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=ppp/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L,xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp),xxtmp);}
  return xxtmp;
}

long double fiter2funcoo0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==1) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp+1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L,xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp),xxtmp);}
  return xxtmp;
}

long double fiter2funcoo1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==3) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp-1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp-0.5L,xxtmp);}
  else {smoothen(xxtmp, xxtmp);}
  return xxtmp;
}

long double fiter2funceo0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  qf=qqq/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp+1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp), xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L, xxtmp);}
  return xxtmp;
}

long double fiter2funceo1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  qf=qqq/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp-1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp,xxtmp);}
  else {smoothen(xxtmp-0.5L, xxtmp);}
  return xxtmp;
}

void selectfiterfuncs(int sqpi, int ssln) 
{
      /* function selection for all possible q/p-combos */
      if (2*(pp[sqpi]/2)==pp[sqpi]) {            /* p even (-> q odd) */
	  if (ssln<=2) fiterfunc=&fiter2funcoe0;    /* I_0^i -> I_0^(1-i) after p/2, (q+-1)/2 */
	  else fiterfunc=&fiter2funcoe1;            /* I_1^i -> I_1^(i-1) after p/2, (q+-1)/2 */
      }
      else {                               /* p odd */
	  if (ssln<=2) {
	      if (2*(qq[sqpi]/2)==qq[sqpi])      /* q even */
		  fiterfunc=&fiter2funceo0;        /* I_0^i -> I_1^i after (p+1)/2, q/2 */
	      else fiterfunc=&fiter2funcoo0;       /* I_0^i -> I_1^(1-i) after (p+1)/2, (q+-1)/2 */
	  }
	  else {
	      if (2*(qq[sqpi]/2)==qq[sqpi])      /* q even */
		  fiterfunc=&fiter2funceo1;        /* I_0^1 -> I_1^1 after (p+1)/2, q/2 */
	      else fiterfunc=&fiter2funcoo1;       /* I_0^1 -> I_1^0 after (p+1)/2, (q+1)/2 */
	  }
      }
}

void mkstr(char *cnummer, unsigned long int nummer)
{
    unsigned long int tmp=nummer;
    char tmpstr[160], nostr[160];

    strcpy(tmpstr,"");
    while (tmp>0) {
	switch (tmp-10*(tmp/10)) {
	    case 1: {strcpy(nostr,"1"); break;}
	    case 2: {strcpy(nostr,"2"); break;}
	    case 3: {strcpy(nostr,"3"); break;}
	    case 4: {strcpy(nostr,"4"); break;}
	    case 5: {strcpy(nostr,"5"); break;}
	    case 6: {strcpy(nostr,"6"); break;}
	    case 7: {strcpy(nostr,"7"); break;}
	    case 8: {strcpy(nostr,"8"); break;}
	    case 9: {strcpy(nostr,"9"); break;}
	    case 0: {strcpy(nostr,"0"); break;}
	}
	strcat(nostr,tmpstr);
	strcpy(tmpstr,nostr);
      	tmp/=10;
    }
    strcpy(cnummer,tmpstr);
}

void zbrak2(long double (*fx)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
	   long double x1, long double x2, int nn, 
	   long double xb1[], long double xb2[], int *nb)
{
  int nbb, i;
  long double x, fp, fc, dx;
  
  nbb=0;
  dx=(x2-x1)/nn;
  fp=(*fx)(par0, x=x1, par1, par2, par3);
  for (i=1;i<=nn;i++) {
    fc=(*fx)(par0, x += dx, par1, par2, par3);
    if (fc*fp <= 0.0) {
      xb1[++nbb]=x-dx;
      xb2[nbb]=x;
      if (*nb==nbb) return;
    }
    fp=fc;
  }
  *nb = nbb;
}

int zbrac2(long double (*func)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
	   long double minbrac, long double maxbrac, long double *x1, long double *x2) 
{
    int j, NTRY=50;
    long double f1, f2, FACTOR=1.6, TOOBIG=0.1L/LDBL_EPSILON;

    if (*x1 == *x2) (*x2 += 10.0L*LDBL_EPSILON);
    if (*x1 > *x2) { f1=*x1; *x2=*x1; *x1=f1; }
    f1=(*func)(par0, *x1, par1, par2, par3);
    f2=(*func)(par0, *x2, par1, par2, par3);
    for (j=1;j<=NTRY;j++) {
	if (f1*f2 < 0.0) return 1;
	if (fabsl(f1) < fabsl(f2) && *x1+FACTOR*(*x1 - *x2)>minbrac)
	    f1=(*func)(par0, *x1 += FACTOR*(*x1 - *x2), par1, par2, par3);
	else
	    f2=(*func)(par0, *x2 += FACTOR*(*x2 - *x1), par1, par2, par3);
	if (fabsl(*x2)>TOOBIG || (*x2>maxbrac)) break;
    }
    return 0;
}


long double zbrent2(long double (*func)(long double, long double, 
				       int, unsigned long int, unsigned long int), 
		   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
		   long double x1, long double x2)
{
    long double EPS=LDBL_EPSILON, ztol=LDBL_EPSILON;
    int ITMAX=100;

    int i;
    long double a=x1,b=x2,c=x2,d,e,min1,min2;
    long double fa, fb, fc,p,q,r,s,tol1,xm;

    fa = (*func)(par0, a, par1, par2, par3); fb = (*func)(par0, b, par1, par2, par3);
    if ((fa > 0.0L && fb > 0.0L) || (fa < 0.0L && fb < 0.0L)) return ERROR;
    fc=fb;
    for (i=1;i<=ITMAX;i++) {
	if ((fb > 0.0L && fc > 0.0L) || (fb < 0.0L && fc < 0.0L)) {
	    c=a;
	    fc=fa;
	    e=d=b-a;
	}
	if (fabsl(fc) < fabsl(fb)) {
	    a=b;
	    b=c;
	    c=a;
	    fa=fb;
	    fb=fc;
	    fc=fa;
	}
	tol1=2.0L*EPS*fabsl(b)+0.5L*ztol;
	xm=0.5L*(c-b);
	if (fabsl(xm) <= tol1 || fb == 0.0L) return b;
	if (fabsl(e) >= tol1 && fabsl(fa) > fabsl(fb)) {
	    s=fb/fa;
	    if (a == c) {
		p=2.0L*xm*s;
		q=1.0L-s;
	    } else {
		q=fa/fc;
		r=fb/fc;
		p=s*(2.0L*xm*q*(q-r)-(b-a)*(r-1.0L));
		q=(q-1.0L)*(r-1.0L)*(s-1.0L);
	    }
	    if (p > 0.0L) q = -q;
	    p=fabsl(p);
	    min1=3.0L*xm*q-fabsl(tol1*q);
	    min2=fabsl(e*q);
	    if (2.0L*p < (min1 <min2 ? min1 : min2)) {
		e=d;
		d=p/q;
	    } else {
		d=xm;
		e=d;
	    }
	} else {
	    d=xm;
	    e=d;
	}
	a=b;
	fa=fb;
	if (fabsl(d) > tol1)
	    b += d;
	else
	    b += (xm > 0.0L ? fabsl(tol1) : -fabsl(tol1));
	fb=(*func)(par0, b, par1, par2, par3);
    }
    return ERROR;
}


int qsortcompare(const void *a, const void *b)
{
  return (((long double *)a)[0]<((long double *)b)[0]? 
	  1 : (((long double *)a)[0]>((long double *)b)[0] ? -1 : 0));
}
