#include "taylor.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

MY_FLOAT mya[3], myl, myomega, mym, myg=1.0;

MY_FLOAT stopT, h, outsteps, dstrobeT, dtheta, dptheta, dpphi;
MY_FLOAT thph0[8], xyz[6];
int      loops, order=2;
char iofilename[160], sosfilename[160], iofilenamebase[80];

void     makeinitials(void);
void     xyz2tp(MY_FLOAT *, MY_FLOAT *);
void     xyz2ex(MY_FLOAT *, MY_FLOAT *);
void     tp2xyz(MY_FLOAT *, MY_FLOAT *, MY_FLOAT);
void     ex2xyz(MY_FLOAT *, MY_FLOAT *, MY_FLOAT);
void     tp2ex(MY_FLOAT *, MY_FLOAT *);
void     ex2tp(MY_FLOAT *, MY_FLOAT *);
MY_FLOAT xbar(MY_FLOAT, int);
MY_FLOAT xbardot(MY_FLOAT, int);
MY_FLOAT en(MY_FLOAT, MY_FLOAT *, int);
MY_FLOAT ptheta(MY_FLOAT, MY_FLOAT *);
MY_FLOAT pphi(MY_FLOAT, MY_FLOAT *);
MY_FLOAT peta(MY_FLOAT, MY_FLOAT *);
MY_FLOAT pxi(MY_FLOAT, MY_FLOAT *);

main(int argc, char *argv[])
{
  double    log10tolerance=-10, log10rtolerance=-10;
  int       o=0, i=0, j=0, jp=0, k=0, itmp=0, direction = 1;
  MY_FLOAT  startT=0.0, nextT, s, en0;
  MY_FLOAT  thph[8], tpo[8], tp[8];
  FILE      *fl, *fsos;

  const MY_FLOAT pi=4.0*atan(1.0);

  strcpy(iofilenamebase,argv[1]);
  makeinitials();

  /* preparing output files */
  fl=fopen(iofilename, "w");
  fprintf(fl, "# a:        %20.15f %20.15f %20.15f \n", mya[0], mya[1], mya[2]);
  fprintf(fl, "# q:        %20.15f %20.15f %20.15f \n", xyz[0], xyz[1], xyz[2]);
  fprintf(fl, "# qdot:     %20.15f %20.15f %20.15f \n", xyz[3], xyz[4], xyz[5]);
  fprintf(fl, "# l:        %20.15f \n", myl);
  fprintf(fl, "# omega:    %20.15f \n", myomega);
  fprintf(fl, "# m:        %20.15f \n", mym);
  fprintf(fl, "# h:        %20.15f \n", h);
  fprintf(fl, "# outsteps: %20.15f \n", outsteps);
  fprintf(fl, "# dstrobeT: %20.15f \n", dstrobeT);
  fprintf(fl, "# stopT:    %20.10f \n", stopT);
  
  fsos=fopen(sosfilename, "w");
  fprintf(fsos, "# a:        %20.15f %20.15f %20.15f \n", mya[0], mya[1], mya[2]);
  fprintf(fsos, "# q:        %20.15f %20.15f %20.15f \n", xyz[0], xyz[1], xyz[2]);
  fprintf(fsos, "# qdot:     %20.15f %20.15f %20.15f \n", xyz[3], xyz[4], xyz[5]);
  fprintf(fsos, "# l:        %20.15f \n", myl);
  fprintf(fsos, "# omega:    %20.15f \n", myomega);
  fprintf(fsos, "# m:        %20.15f \n", mym);
  fprintf(fsos, "# h:        %20.15f \n", h);
  fprintf(fsos, "# outsteps: %20.15f \n", outsteps);
  fprintf(fsos, "# dstrobeT: %20.15f \n", dstrobeT);
  fprintf(fsos, "# stopT:    %20.10f \n", stopT);

  printf("----------- P r o g r e s s ------------\n");

  for (k=0; k<loops; k++) {        /* several initial conditions */

    /* initializations */
    thph[0] = thph0[0]+dtheta*k;
    thph[2] = thph0[2];
    /* note: pphi & ptheta are varied (or not), THEN recalc thetadot & phidot */
    thph[1] = (ptheta(0.0,&thph0[0])+k*dptheta)/mym/myl/myl -
      (xbardot(0.0,0)*cos(thph[0])*cos(thph[2])+
       xbardot(0.0,1)*cos(thph[0])*sin(thph[2])-
       xbardot(0.0,2)*sin(thph[0]))/myl;
    thph[3] = ((pphi(0.0,&thph0[0])+k*dpphi)/mym/myl/sin(thph[0])-
	       (-xbardot(0.0,0)*sin(thph[2])+
		xbardot(0.0,1)*cos(thph[2])))/myl/sin(thph[0]);
    tp2ex(&thph[4],&thph[0]);
    if (thph[0]>pi-pi/6.0 || thph[0]<pi/6.0) o=4; else o=0;
    
    startT = 0.0; j=0; jp=0;
    en0 = en (startT, &thph[o], o); 
    for (i=0; i<8; i++) tpo[i] = thph[i];

    /* prepare output files */
    fprintf(fl, "\n# theta:    %20.15f    ptheta:   %20.15f \n", thph[0], ptheta(startT,&thph[0]));
    fprintf(fl, "# phi:      %20.15f    pphi:     %20.15f \n", thph[2], pphi(startT,&thph[0]));
    fprintf(fl, "# en0:      %20.15f\n# \n", en0);
    fprintf(fl, "# %10s%10s%10s%10s%10s%10s%10s%10s%10s%10s%10s%14s%2s\n",
	    "time", "x", "y", "z", "px", "py", "pz", 
	    "theta", "phi", "ptheta", "pphi", "delta E","o");
    
    fprintf(fsos, "\n# theta:    %20.15f    ptheta:   %20.15f \n", thph[0], ptheta(startT,&thph[0]));
    fprintf(fsos, "# phi:      %20.15f    pphi:     %20.15f \n", thph[2], pphi(startT,&thph[0]));
    fprintf(fsos, "# en0:      %20.15f \n# \n", en0);
    fprintf(fsos, "# %14s %16s %16s %16s %16s %12s %12s%2s\n",
	    "time", "time+sh", "theta", "ptheta", "pphi", "delta E", "s","o");
    if ((thph[2]==0.0 && pphi(startT, &thph[0])>=0.0) ||
	(thph[6]==-pi/2 && pxi(startT, &thph[4])>=0.0))
      fprintf(fsos, "%16.6f %16.6f %16.6f %16.6f %16.6f %12.6g %12.6g%2d\n",
	      startT, startT, thph[0], ptheta(startT,&thph[0]), pphi(startT,&thph[0]),
	      en(startT,&thph[o],o)-en0,0.0,o);


    /* main loop */
    do {
      
      /* modulo 2 pi */
      for (i=0; i<8; i+=2) {
	while (thph[i]> pi) thph[i]=thph[i]-2.0*pi;
	while (thph[i]< -pi) thph[i]=thph[i]+2.0*pi;
      }
      
      /* output of all results in time intervals of outsteps */
      if (startT+h/1000.0>=outsteps*j) {
	j++;
	if (o==4) {
	  ex2xyz(&xyz[0],&thph[4],startT);
	  ex2tp(&thph[0],&thph[4]);
	}
	else tp2xyz(&xyz[0],&thph[0],startT);
	
	fprintf(fl, "%12.6f%10.6f%10.6f%10.6f%10.6f%10.6f%10.6f",
		startT, xyz[0], xyz[1], xyz[2], 
		xyz[3], xyz[4], xyz[5]);
	fprintf(fl, "%10.6f%10.6f%10.6f%10.6f%14.6g%2d\n",
		thph[0], thph[2], ptheta(startT, &thph[0]), 
		pphi(startT,&thph[0]), en(startT,&thph[o],o)-en0,o);
      }

      /* poincare sos output whenever time~mult. of T and phi=0 plane is crossed */
 
      if ((o==0 && tpo[2]<0.0 && thph[2]>=0.0) ||         /* xz-plane: phi=0     */
	  (o==4 && tpo[6]<-pi/2.0 && thph[6]>-pi/2.0))    /* xz-plane: xi=+-pi/2 */
      {
	s=(-pi/8.0*o-tpo[o+2])/(thph[o+2]-tpo[o+2]);
	if (startT+(s-1.0)*h-dstrobeT/2.0<=jp*2.0*pi/myomega && 
	    startT+(s-1.0)*h+dstrobeT/2.0>jp*2.0*pi/myomega) {

	  for (i=0; i<4; i++) tp[o+i] = tpo[o+i] + (thph[o+i]-tpo[o+i])*s;
	  if (o==4) ex2tp(&tp[0],&tp[4]);

	  fprintf(fsos, "%16.6f %16.6f %16.6f %16.6f %16.6f %12.6g %12.6g%2d\n",
		  startT, startT+(s-1.0)*h, 
		  tp[0], ptheta(startT,&tp[0]), pphi(startT,&tp[0]),
		  en(startT,&tp[o],o)-en0,s,o);
	}
      }
      if (startT>=jp*2.0*pi/myomega) jp++;

      /* save old results */
      for (i=0; i<4; i++) tpo[o+i]=thph[o+i];

      /* next integration */
      nextT=h;
      if (o==0) taylor_step_pend(&startT, &thph[o], direction, 0, log10tolerance, 
				 log10rtolerance, &stopT, &nextT, &order);
      else taylor_step_pend2(&startT, &thph[o], direction, 0, log10tolerance, 
			     log10rtolerance, &stopT, &nextT, &order);

      if ((o==0) && (thph[0]>pi-pi/6.0 || thph[0]<pi/6.0)) {
	o=4;
	tp2ex(&tpo[4], &tpo[0]);
	startT-=h;
	nextT=h;
        for (i=0; i<4; i++) thph[o+i]=tpo[o+i];
	taylor_step_pend2(&startT, &thph[o], direction, 0, log10tolerance, 
			  log10rtolerance, &stopT, &nextT, &order);
      }
      else if ((o==4) && (fabs(sin(thph[4])*sin(thph[6]))<0.8660254038)) {
	o=0;
	ex2tp(&tpo[0], &tpo[4]);
	startT-=h;
	nextT=h;
        for (i=0; i<4; i++) thph[o+i]=tpo[o+i];
	taylor_step_pend(&startT, &thph[o], direction, 0, log10tolerance, 
			  log10rtolerance, &stopT, &nextT, &order);
      }


      /* progress report */
      if (floor(40.0*(startT/stopT+ (double)(k))/((double)loops))
	  >floor(40.0*((startT-nextT)/stopT+ (double)(k))/((double)loops))) 
	{
	  printf("*"); fflush(stdout);
	}
    } while(startT < stopT); /* while stopT not reached */ 
  }

  printf("\n");
  fclose(fl);
  fclose(fsos);
  exit(0); 
} 


void makeinitials(void)      /* read and adjust initial values */
{
  int       i=0;
  FILE      *fl;
  char      dummy[80];

  const MY_FLOAT pi=4.0*atan(1.0);

  strcpy(iofilename,iofilenamebase);
  strcat(iofilename,".in");

  /* read initials */
  if ((fl=fopen(iofilename, "r")) == NULL) {
    fprintf(stderr, "*** error opering in-file  %s  ***\n", iofilename);
    exit(1);
  }

  fscanf(fl,"%lf %lf %lf", &mya[0], &mya[1], &mya[2]); /* a                  */
  fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &myl);     fgets(dummy, 80, fl);    /* l^2                */
  fscanf(fl,"%lf", &myomega); fgets(dummy, 80, fl);    /* omega^2            */
  fscanf(fl,"%lf", &mym);     fgets(dummy, 80, fl);    /* m                  */
  fscanf(fl,"%lf %lf %lf", &xyz[0], &xyz[1], &xyz[2]); /* x^2                */
  fgets(dummy, 80, fl);
  fscanf(fl,"%lf %lf %lf", &xyz[3], &xyz[4], &xyz[5]); /* p^2                */
  fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &h);       fgets(dummy, 80, fl);    /* maximal stepsize /(2 pi/om)  */
  fscanf(fl,"%lf", &outsteps);fgets(dummy, 80, fl);    /* div. by 2 pi/omega */
  fscanf(fl,"%lf", &dstrobeT);fgets(dummy, 80, fl);    /* div. by 2 pi/omega */
  fscanf(fl,"%ld", &order);   fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &stopT);   fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &dtheta);  fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &dptheta); fgets(dummy, 80, fl);
  fscanf(fl,"%lf", &dpphi);   fgets(dummy, 80, fl);
  fscanf(fl,"%ld", &loops);   fgets(dummy, 80, fl);
  fclose(fl);

  strcpy(iofilename,iofilenamebase);
  strcat(iofilename,".out");
  strcpy(sosfilename,iofilenamebase);
  strcat(sosfilename,".psos.out");

  /* adjust initials */
  myl = sqrt(myl);
  myomega = sqrt(myomega);
  for (i=0; i<6; i++) 
    if (fabs(xyz[i])>0.0) xyz[i] = xyz[i]/fabs(xyz[i])*sqrt(fabs(xyz[i])); 	   
  xyz2tp(&thph0[0],&xyz[0]);
  xyz2ex(&thph0[4],&xyz[0]);  
  h = h * 2.0 * pi/myomega;
  outsteps = outsteps * 2.0 * pi/myomega;
  dstrobeT = dstrobeT * 2.0 * pi/myomega;
}


void xyz2tp(MY_FLOAT *tp, MY_FLOAT *qq)
{
  MY_FLOAT rd;
  const MY_FLOAT pi=4.0*atan(1.0);
  rd=myl*myl-qq[2]*qq[2];
  
  if (rd>0.0) {
    tp[0] = pi/2.0-atan(qq[2]/sqrt(rd));
    tp[2] = (qq[0]==0) ? 
      ( (qq[1]==0.0) ? 0.0 : pi/2.0*qq[1]/fabs(qq[1]) ): atan2(qq[1], qq[0]);
    tp[1] = sqrt(rd)/myl/myl * ((qq[0]*qq[3]+qq[1]*qq[4])/rd*qq[2]-qq[5]);
    tp[3] = (qq[0]*qq[4]-qq[1]*qq[3])/rd;
  }
  else {
    tp[0]=pi/2.0*(1.0-qq[2]*qq[2]/myl/myl);
    tp[2] = (qq[0]==0) ? 
      ( (qq[1]==0.0) ? 0.0 : pi/2.0*qq[1]/fabs(qq[1]) ): atan2(qq[1], qq[0]);
    tp[1]=0.0;
    tp[3]=0.0;
  }

}


void xyz2ex(MY_FLOAT *ex, MY_FLOAT *qq)
{
  MY_FLOAT rd;
  const MY_FLOAT pi=4.0*atan(1.0);
  rd=myl*myl-qq[0]*qq[0];
  
  if (rd>0.0) {
    ex[0] = pi/2.0-atan(qq[0]/sqrt(rd));
    ex[2] = (qq[1]==0) ? 
      ( (qq[2]==0.0) ? 0.0 : pi/2.0*qq[2]/fabs(qq[2]) ): atan2(qq[2], qq[1]);
    ex[1] = sqrt(rd)/myl/myl * ((qq[1]*qq[4]+qq[2]*qq[5])/rd*qq[0]-qq[3]);
    ex[3] = (qq[1]*qq[5]-qq[2]*qq[4])/rd;
  }
  else {
    ex[0]=pi/2.0*(1.0-qq[0]*qq[0]/myl/myl);
    ex[2] = (qq[1]==0) ? 
      ( (qq[2]==0.0) ? 0.0 : pi/2.0*qq[2]/fabs(qq[2]) ): atan2(qq[2], qq[1]);
    ex[1]=0.0;
    ex[3]=0.0;
  }

}


void tp2xyz(MY_FLOAT *qq, MY_FLOAT *tp, MY_FLOAT t)
{
  MY_FLOAT rd,pth,pph;

  qq[0]=myl*sin(tp[0])*cos(tp[2]);
  qq[1]=myl*sin(tp[0])*sin(tp[2]);
  qq[2]=myl*cos(tp[0]);

  pth=ptheta(t,&tp[0]);
  pph=pphi(t,&tp[0]);
  rd=myl*myl-qq[2]*qq[2];
  if (rd>0.0) {
    qq[3]=(mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))+
	   pth*qq[2]/sqrt(rd))*qq[0]/myl/myl-pph*qq[1]/rd;
    qq[4]=(mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))+
	   pth*qq[2]/sqrt(rd))*qq[1]/myl/myl+pph*qq[0]/rd;
    qq[5]=mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))*
      qq[2]/myl/myl-sqrt(rd)*pth/myl/myl;
  }
  else {
    qq[3]=0.0;
    qq[4]=0.0;
    qq[5]=0.0;
  }
}

void ex2xyz(MY_FLOAT *qq, MY_FLOAT *ex, MY_FLOAT t)
{
  MY_FLOAT rd,pe,px;

  qq[0]=myl*cos(ex[0]);
  qq[1]=myl*sin(ex[0])*cos(ex[2]);
  qq[2]=myl*sin(ex[0])*sin(ex[2]);

  pe=peta(t,&ex[0]);
  px=pxi(t,&ex[0]);
  rd=myl*myl-qq[0]*qq[0];
  if (rd>0.0) {
    qq[3]=mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))*
      qq[0]/myl/myl-sqrt(rd)*pe/myl/myl;
    qq[4]=(mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))+
	   pe*qq[0]/sqrt(rd))*qq[1]/myl/myl-px*qq[2]/rd;
    qq[5]=(mym*(qq[0]*xbardot(t,0)+qq[1]*xbardot(t,1)+qq[2]*xbardot(t,2))+
	   pe*qq[0]/sqrt(rd))*qq[2]/myl/myl+px*qq[1]/rd;
  }
  else {
    qq[3]=0.0;
    qq[4]=0.0;
    qq[5]=0.0;
  }  
}

void tp2ex(MY_FLOAT *ex, MY_FLOAT *tp)
{
  MY_FLOAT rd;
  const MY_FLOAT pi=4.0*atan(1.0);
  rd=1.0-sin(tp[0])*sin(tp[0])*cos(tp[2])*cos(tp[2]);
  if (rd>0.0) {
    ex[0]=pi/2.0-atan(sin(tp[0])*cos(tp[2])/sqrt(rd));
    ex[2]= (sin(tp[0])*sin(tp[2])==0) ?
      ( (cos(tp[0])==0.0) ? 0.0 : pi/2.0*cos(tp[0])/fabs(cos(tp[0])) ):
      atan2(cos(tp[0]),sin(tp[0])*sin(tp[2]));
    ex[1]=(-tp[1]*cos(tp[0])*cos(tp[2])+tp[3]*sin(tp[0])*sin(tp[2]))/sqrt(rd);
    ex[3]=(-tp[1]*sin(tp[2])-tp[3]*sin(tp[0])*cos(tp[0])*cos(tp[2]))/rd;
  }
  else {
    ex[0]=pi/2.0*(1.0-sin(tp[0])*cos(tp[2]));
    ex[2]= (sin(tp[0])*sin(tp[2])==0) ?
      ( (cos(tp[0])==0.0) ? 0.0 : pi/2.0*cos(tp[0])/fabs(cos(tp[0])) ):
      atan2(cos(tp[0]),sin(tp[0])*sin(tp[2]));
    ex[1]=0.0;
    ex[3]=0.0;
  }
}

void ex2tp(MY_FLOAT *tp, MY_FLOAT *ex)
{
  MY_FLOAT rd;
  const MY_FLOAT pi=4.0*atan(1.0);
  rd=1.0-sin(ex[0])*sin(ex[0])*sin(ex[2])*sin(ex[2]);
  if (rd>0.0) {
    tp[0]=pi/2.0-atan(sin(ex[0])*sin(ex[2])/sqrt(rd));
    tp[2]= (cos(ex[0])==0) ? 
      ( (sin(ex[0])*cos(ex[2])==0.0) ? 0.0 : 
	pi/2.0*sin(ex[0])*cos(ex[2])/fabs(sin(ex[0])*cos(ex[2])) ):
      atan2(sin(ex[0])*cos(ex[2]),cos(ex[0]));
    tp[1]=(-ex[1]*cos(ex[0])*sin(ex[2])-ex[3]*sin(ex[0])*cos(ex[2]))/sqrt(rd);
    tp[3]=(ex[1]*cos(ex[2])-ex[3]*sin(ex[0])*cos(ex[0])*sin(ex[2]))/rd;
  }
  else {
    tp[0]=pi/2.0*(1.0-sin(ex[0])*sin(ex[2]));
    tp[2]= (cos(ex[0])==0) ? 
      ( (sin(ex[0])*cos(ex[2])==0.0) ? 0.0 : 
	pi/2.0*sin(ex[0])*cos(ex[2])/fabs(sin(ex[0])*cos(ex[2])) ):
      atan2(sin(ex[0])*cos(ex[2]),cos(ex[0]));
    tp[1]=0.0;
    tp[3]=0.0;
  }
}

MY_FLOAT xbar(MY_FLOAT t, int comp)       /* position of suspension -> acc. see pendel.eq1 */
{
  return mya[comp]*myl*sin(myomega*t);
}


MY_FLOAT xbardot(MY_FLOAT t, int comp)    /* velocity of suspension -> acc. see pendel.eq1 */ 
{
  return mya[comp]*myl*myomega*cos(myomega*t);
}


MY_FLOAT en(MY_FLOAT t, MY_FLOAT *tp, int ord)     /* energy */
{
  int i;
  MY_FLOAT xbd[3], xbds=0;
  
  for (i=0; i<3; i++) {
    xbd[i] = xbardot(t,i);
    xbds += xbd[i]*xbd[i];
  }

  if (ord==0)
    return mym/2.0*(xbds + myl*myl*tp[1]*tp[1] + 
		    myl*myl*tp[3]*tp[3]*sin(tp[0])*sin(tp[0])) +
      mym*myl*tp[1]*(xbd[0]*cos(tp[0])*cos(tp[2]) + xbd[1]*cos(tp[0])*sin(tp[2]) -
		     xbd[2]*sin(tp[0]))+
      mym*myl*tp[3]*sin(tp[0])*(-xbd[0]*sin(tp[2])+xbd[1]*cos(tp[2]))+
      mym*myg*(xbar(t,2)+myl*cos(tp[0]));
  else
    return mym/2.0*(xbds + myl*myl*tp[1]*tp[1] + 
		    myl*myl*tp[3]*tp[3]*sin(tp[0])*sin(tp[0])) +
      mym*myl*tp[1]*(xbd[1]*cos(tp[0])*cos(tp[2]) + xbd[2]*cos(tp[0])*sin(tp[2]) -
		     xbd[0]*sin(tp[0]))+
      mym*myl*tp[3]*sin(tp[0])*(-xbd[1]*sin(tp[2])+xbd[2]*cos(tp[2]))+
      mym*myg*(xbar(t,0)+myl*sin(tp[0])*sin(tp[2]));
}


MY_FLOAT ptheta(MY_FLOAT t, MY_FLOAT *tp)  /* momentum conj. to theta */
{
  return mym*myl*myl*tp[1] + mym*myl* (xbardot(t,0)*cos(tp[0])*cos(tp[2])+
				       xbardot(t,1)*cos(tp[0])*sin(tp[2])- 
				       xbardot(t,2)*sin(tp[0]));
}


MY_FLOAT pphi(MY_FLOAT t, MY_FLOAT *tp)    /* momentum conj. to phi */
{
  return mym*myl*myl*tp[3]*sin(tp[0])*sin(tp[0])+
    mym*myl*sin(tp[0])* (-xbardot(t,0)*sin(tp[2])+xbardot(t,1)*cos(tp[2]));
}


MY_FLOAT peta(MY_FLOAT t, MY_FLOAT *ex)    /* momentum conj. to eta */
{
  return mym*myl*myl*ex[1] + mym*myl* (xbardot(t,1)*cos(ex[0])*cos(ex[2])+
				       xbardot(t,2)*cos(ex[0])*sin(ex[2])- 
				       xbardot(t,0)*sin(ex[0]));
}


MY_FLOAT pxi(MY_FLOAT t, MY_FLOAT *ex)     /* momentum conj. to xi */
{
  return mym*myl*myl*ex[3]*sin(ex[0])*sin(ex[0])+
    mym*myl*sin(ex[0])* (-xbardot(t,1)*sin(ex[2])+xbardot(t,2)*cos(ex[2]));
}
