#include <stdio.h>
#include <math.h>
#include <string.h>
#include <float.h>
#include <stdlib.h>

#define NL(flnm,dumch) {do dumch=fgetc(flnm); while (dumch!='\n' && dumch!=EOF);}
#define ntmapstep(x,y,a,b) {y-=(b)*sinl(pi2*(x)); x+=(a)-(a)*(y)*(y);}
#define smoothen(x,y) {if ((x)>0.0L) y=((x)/(1.0L+(x))); else y=((x)/(1.0L-(x)));}
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d); /* for mnbrac & brent */
#define SIGN(a,b) ((b) >= 0.0 ? fabsl(a) : -fabsl(a))
#define FMAX(a,b) ((a) > (b) ? (a) : (b))
#define IPOW3(a) ((int)(pow(3.0L, (long double)(a))+0.5L)) 
#define ERROR 123.456789L
#define UNDEFINED -123456.789L

const long double SMALL=1.0e-10L;
const int SUPPERROR=15;

void makeinitials(void);
void prepoutfiles(void);
void maincalc(void);
void selectfiterfuncs(int, int);
void invquadfit(long double, long double, long double, long double *, long double *);
long double quadfit(long double, long double [6]);
void invcubfit(long double, long double, long double, long double *, long double *);
long double cubfit(long double, long double [6]);
long double qcfit(long double, long double [6]);
long double rootybfunc(long double, int, unsigned long int, unsigned long int, long double, long double, 
		       long double, long double);
long double rootbyfunc(long double, int, unsigned long int, unsigned long int, long double, long double);
long double findx0(long double, long double, int);
long double fiter2funcoe0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoo0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funceo0(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoe1(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funcoo1(long double, long double, int, unsigned long int, unsigned long int);
long double fiter2funceo1(long double, long double, int, unsigned long int, unsigned long int);
void zbrak(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double, int, unsigned long int, unsigned long int, 
	   long double, long double, int, long double [], long double [], int *);
int zbrac2(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double, int, unsigned long int, unsigned long int, long double, long double,
	   long double *, long double *);
long double zbrent(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
		   long double, int, unsigned long int, unsigned long int, long double, long double);
long double zbrent2(long double (*)(long double, long double, int, unsigned long int, unsigned long int), 
		   long double, int, unsigned long int, unsigned long int, long double, long double);
void mnbrac(long double *, long double *, long double *, 
	    long double (*)(long double, int, unsigned long int, unsigned long int, long double, long double), 
	    int, unsigned long int, unsigned long int, long double, long double);
int mymnbrac(long double *, long double *, long double *, long double,
	    long double (*)(long double, int, unsigned long int, unsigned long int, long double, long double), 
	    int, unsigned long int, unsigned long int, long double, long double);
long double brent(long double, long double, long double, 
	    long double (*)(long double, int, unsigned long int, unsigned long int, long double, long double), 
		  int, unsigned long int, unsigned long int, long double, long double, long double *);
void mkstr(char *, unsigned long int);
int qsortcompare(const void *, const void *);

long double pi2, minmax;
long double omega[50], a0, a00, a0min, a0max, a0step, b0, b00, ymin, ymax, y00, bxx;
long double yguess, bguess, INTERVALS, bsharp;
int sln, slnmin, slnmax, qpi, qp0, qpx, bres;
unsigned long int qq[50], pp[50], qqh[50], pph[50];
char iterfilename[160], iterfilenamebase[80];
long double braku[3], brakl[3];

FILE *fl1, *fl2, *fl3, *fl;
long double (*fiterfunc)(long double, long double, int, unsigned long int, unsigned long int);

int main(int argc, char* argv[])
{
  /* read input */
  if ( (fl=fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "*** error opering in-file  %s  ***\n", argv[1]);
    exit(1);
  }

  makeinitials();
  fclose(fl);
  
  for (sln=slnmin;sln<=slnmax;sln++) {
      
      /* prepare output files */
      prepoutfiles();
      
      maincalc();
      
  }
  
  fclose(fl1);
  fclose(fl2);
  fclose(fl3);
  exit(0); 
} 


void makeinitials(void)      /* read and adjust initial values */
{
  char dummy;
  int i, iintervals, cfn, cfp, cfa[50];
  long double tmp;
  
  pi2=8.0L*atanl(1.0L);

  /* read initials */
  fscanf(fl,"%d", &sln); NL(fl, dummy);                /* shinohara point */
  fscanf(fl,"%Lf %Lf", &a00, &a0max);  NL(fl, dummy);  /* a0-range*/
  fscanf(fl,"%Lf", &a0step);  NL(fl, dummy);           /* a0-step */
  fscanf(fl,"%Lf %Lf", &ymin, &ymax);  NL(fl, dummy);  /* start y-range */
  fscanf(fl,"%Lf", &yguess);  NL(fl, dummy);           /* guess for initial y */
  fscanf(fl,"%Lf %Lf", &b00, &bxx);  NL(fl, dummy);    /* start b-range */
  fscanf(fl,"%Lf", &bguess);  NL(fl, dummy);           /* guess for initial b */
  fscanf(fl,"%d", &bres);  NL(fl, dummy);              /* # of brackets if initial b-bracket fails */
  fscanf(fl,"%d", &iintervals);  NL(fl, dummy);        /* # of mymnbrak-"intervals" */
  fscanf(fl,"%s", iterfilenamebase); NL(fl, dummy);    /* outfilename */
  fscanf(fl,"%Lf", &minmax);  NL(fl, dummy);           /* factor for rootbyfunc +-1 -> min or max */
  fscanf(fl,"%d %d", &qp0, &qpx);  NL(fl, dummy);      /* start and end number of p/q values in FLS*/
  fscanf(fl,"%d %d", &cfn, &cfp); NL(fl, dummy);       /* number of cf coeff. and period */
  for (i=0;i<cfn;i++) fscanf(fl,"%d ", &cfa[i]); NL(fl, dummy); 

  INTERVALS=(long double)(iintervals);
  if (sln==0) {slnmin=1; slnmax=4;}
  else {slnmin=sln; slnmax=sln;}
  if (bxx<b00) {tmp=b00; b00=bxx; bxx=tmp;}
  if (bres<1) bres=1;
  if (qp0<1) qp0=1;
  if (qpx<qp0) qpx=qp0;
  if (cfp>cfn) cfp=1;

  for (i=cfn;i<=qpx;i++) cfa[i]=cfa[cfn-cfp+(i-cfn)%cfp];
  qq[0]=1; pp[0]=cfa[0];
  qq[1]=cfa[1]; pp[1]=pp[0]*cfa[1]+1;
  qqh[0]=1; pph[0]=cfa[0];
  qqh[1]=cfa[1]; pph[1]=pp[0]*cfa[1]+1;
  omega[0]=((long double)(qq[0]))/((long double)(pp[0]));
  omega[1]=((long double)(qq[1]))/((long double)(pp[1]));
  a0min=a00; b0=b00;
  if (a00<=omega[1]) { a0min=omega[1]; b0=0.0L; }
  for (i=2;i<=qpx;i++) {
      qq[i]=cfa[i]*qq[i-1]+qq[i-2];
      pp[i]=cfa[i]*pp[i-1]+pp[i-2];      
      qqh[i]=qq[i-1]+qq[i-2];
      pph[i]=pp[i-1]+pp[i-2];      
      omega[i]=((long double)(qq[i]))/((long double)(pp[i]));
      if (a00<=omega[i]) { a0min=omega[i]; b0=0.0L; }
  }
  if (b0<=0.0L) b0=0.0L; else b0=b00;
  if (minmax<0.0L) minmax=-1.0L; else minmax=1.0L;
}


void prepoutfiles(void)
{
  char ctmp[60];
  int i;
    
  strcpy(iterfilename,iterfilenamebase);
  strcpy(ctmp,""); mkstr(&ctmp[0],qp0); strcat(iterfilename,ctmp);
  strcat(iterfilename,"_");
  strcpy(ctmp,""); mkstr(&ctmp[0],qpx); strcat(iterfilename,ctmp);
  strcat(iterfilename,".s");
  strcpy(ctmp,""); mkstr(&ctmp[0],sln); strcat(iterfilename,ctmp);

  fl1=fopen(iterfilename, "w");
  fprintf(fl1,"# Symm.l.: %2d \n", sln);
  fprintf(fl1,"# a0:      %10.7Lf , ax:      %10.7Lf , astep:   %10.7Lf \n", a0min, a0max, a0step); 
  fprintf(fl1,"# ymin:    %10.7Lf , yguess:  %10.7Lf , ymax:    %10.7Lf \n", ymin, yguess, ymax); 
  fprintf(fl1,"# b0:      %10.7Lf , bguess:  %10.7Lf , bx:      %10.7Lf \n", b0, bguess, bxx); 
  fprintf(fl1,"# bres:    %10d , inters:  %10.0Lf \n", bres, INTERVALS); 
  fprintf(fl1,"# "); 
  for (i=qp0; i<=qpx; i++)   fprintf(fl1,"[%d]=%lu/%lu, ", i, qq[i], pp[i]); 
  fprintf(fl1,"~%.18Lf\n", ((long double)(qq[qpx]))/((long double)(pp[qpx]))); 
  fprintf(fl1,"#%4s%24s%24s%24s%24s%24s%10s%10s%5s\n", 
	  "[n]", "a", "b", "y", "bdiff", "ydiff", "c1", "c2", "ret");
  fprintf(fl1,"# codes: 1 max, -1 min, 2 leaving b-range, -2 entering b-range\n");
  fprintf(fl1,"#        0 pseudo-extr. outside b-range\n");
  fprintf(fl1,"#        (n*1000+m)*100+N boundaries - found after n+m iterations/ bound N extrema \n\n");


  strcpy(iterfilename,iterfilenamebase);
  strcpy(ctmp,""); mkstr(&ctmp[0],qp0); strcat(iterfilename,ctmp);
  strcat(iterfilename,"_");
  strcpy(ctmp,""); mkstr(&ctmp[0],qpx); strcat(iterfilename,ctmp);
  strcat(iterfilename,"quick.s");
  strcpy(ctmp,""); mkstr(&ctmp[0],sln); strcat(iterfilename,ctmp);

  fl2=fopen(iterfilename, "w");
  fprintf(fl2,"# Symm.l.: %2d \n", sln);
  fprintf(fl2,"# a0:      %10.7Lf , ax:      %10.7Lf , astep:   %10.7Lf \n", a0min, a0max, a0step); 
  fprintf(fl2,"# ymin:    %10.7Lf , yguess:  %10.7Lf , ymax:    %10.7Lf \n", ymin, yguess, ymax); 
  fprintf(fl2,"# b0:      %10.7Lf , bguess:  %10.7Lf , bx:      %10.7Lf \n", b0, bguess, bxx); 
  fprintf(fl2,"# bres:    %10d , inters:  %10.0Lf \n", bres, INTERVALS); 
  fprintf(fl2,"# "); 
  for (i=qp0; i<=qpx; i++)   fprintf(fl2,"[%d]=%lu/%lu, ", i, qq[i], pp[i]); 
  fprintf(fl2,"~%.18Lf\n", ((long double)(qq[qpx]))/((long double)(pp[qpx]))); 
  fprintf(fl2,"#%25s%25s%25s\n", "a_crit", "b_crit", "y");

  strcpy(iterfilename,iterfilenamebase);
  strcpy(ctmp,""); mkstr(&ctmp[0],qp0); strcat(iterfilename,ctmp);
  strcat(iterfilename,"_");
  strcpy(ctmp,""); mkstr(&ctmp[0],qpx); strcat(iterfilename,ctmp);
  strcat(iterfilename,"calc.s");
  strcpy(ctmp,""); mkstr(&ctmp[0],sln); strcat(iterfilename,ctmp);

  fl3=fopen(iterfilename, "w");
  fprintf(fl3,"# Symm.l.: %2d \n", sln);
  fprintf(fl3,"# a0:      %10.7Lf , ax:      %10.7Lf , astep:   %10.7Lf \n", a0min, a0max, a0step); 
  fprintf(fl3,"# ymin:    %10.7Lf , yguess:  %10.7Lf , ymax:    %10.7Lf \n", ymin, yguess, ymax); 
  fprintf(fl3,"# b0:      %10.7Lf , bguess:  %10.7Lf , bx:      %10.7Lf \n", b0, bguess, bxx); 
  fprintf(fl3,"# bres:    %10d , inters:  %10.0Lf \n", bres, INTERVALS); 
  fprintf(fl3,"# "); 
  for (i=qp0; i<=qpx; i++)   fprintf(fl3,"[%d]=%lu/%lu, ", i, qq[i], pp[i]); 
  fprintf(fl3,"~%.18Lf\n", ((long double)(qq[qpx]))/((long double)(pp[qpx]))); 
  fprintf(fl3,"#%4s%24s%24s%24s\n", "[n]", "a", "b", "y");

}

void maincalc() 
{
    const long double TINY=10.0L*LDBL_EPSILON;
    long double ybraca, ybracb, ybracc, bbraca, bbracc, sol[50][8];
    long double ymintmp, ymaxtmp, tmp;
    int okbrac, innercurve;

    for (a0=a0min;a0<=a0max; a0+=a0step) {	
    bbraca=b0; bbracc=bxx;

    qpi=qp0; selectfiterfuncs(qpi,sln);
    ybraca=ymin; ybracb=yguess; ybracc=ymax;
    if((okbrac=mymnbrac(&ybraca, &ybracb, &ybracc, INTERVALS, rootbyfunc, 
			sln, qq[qp0], pp[qp0], bbraca, bbracc))>=1)
	sol[qp0][1]=-minmax*brent(ybraca, ybracb, ybracc, rootbyfunc, 
			  sln, qq[qp0], pp[qp0], bbraca, bbracc, &sol[qp0][0]);
    sol[qpi][6]=0.0L;
    sol[qpi][7]=0.0L;

    qpi=qp0+1; selectfiterfuncs(qpi,sln);
    ybraca=ymin; ybracb=sol[qp0][0]; ybracc=ymax;
    if((okbrac=mymnbrac(&ybraca, &ybracb, &ybracc, INTERVALS, rootbyfunc, 
			sln, qq[qp0+1], pp[qp0+1], bbraca, bbracc))>=1)
	sol[qp0+1][1]=-minmax*brent(ybraca, ybracb, ybracc, rootbyfunc, 
			    sln, qq[qp0+1], pp[qp0+1], bbraca, bbracc, &sol[qp0+1][0]);
    
    if (-minmax*sol[qp0][1]>-minmax*sol[qp0+1][1]) {
	bbraca=sol[qp0+1][1]; bbracc=sol[qp0][1]; innercurve=qp0; 
	ybracb=sol[qp0][0]; 
	ybraca=rootybfunc(bbracc, sln, qq[qp0+1], pp[qp0+1], 
			  -1.0L/SMALL, (3.0L*ybracb+ymin)/4.0L, ybracb, (ybracb-ymin)/INTERVALS);
	if (fabsl(ybraca-ERROR)<SMALL) {
	    ybraca=rootybfunc(bbracc, sln, qq[qp0+1], pp[qp0+1], 
			      -1.0L/SMALL, ymin, ybracb, (ybracb-ymin)/INTERVALS/INTERVALS);
	    if (fabsl(ybraca-ERROR)<SMALL) ymintmp=ymin; else ymintmp=ybraca;
	}
	else ymintmp=ybraca;
	ybracc=rootybfunc(bbracc, sln, qq[qp0+1], pp[qp0+1], 
			  ybracb, (3.0*ybracb+ymax)/4.0L, 1.0L/SMALL, (ymax-ybracb)/INTERVALS);
	if (fabsl(ybracc-ERROR)<SMALL) {
	    ybracc=rootybfunc(bbracc, sln, qq[qp0+1], pp[qp0+1], 
			      ybracb, ymax, 1.0L/SMALL, (ymax-ybracb)/INTERVALS/INTERVALS);
	    if (fabsl(ybracc-ERROR)<SMALL) ymaxtmp=ymax; else ymaxtmp=ybracc;
	}
	else ymaxtmp=ybracc;
	
	sol[qp0][2]=0.0L; sol[qp0][3]=0.0L;
	sol[qp0][4]=0.0L; sol[qp0][5]=0.0L;
	invquadfit(sol[qp0][1] - sol[qp0+1][1], ymintmp-sol[qp0+1][0], ymaxtmp-sol[qp0+1][0],
		   &sol[qp0+1][2], &sol[qp0+1][3]);
	invcubfit(sol[qp0][1] - sol[qp0+1][1], ymintmp-sol[qp0+1][0], ymaxtmp-sol[qp0+1][0],
		   &sol[qp0+1][4], &sol[qp0+1][5]);
    }
    else {
	bbraca=sol[qp0][1]; bbracc=sol[qp0+1][1]; innercurve=qp0+1; 
	ybracb=sol[qp0+1][0]; 
	selectfiterfuncs(qp0,sln);
	ybraca=rootybfunc(bbracc, sln, qq[qp0], pp[qp0], 
			  -1.0L/SMALL, (3.0L*ybracb+ymin)/4.0L, ybracb, (ybracb-ymin)/INTERVALS);
	if (fabsl(ybraca-ERROR)<SMALL) ymintmp=ymin; else ymintmp=ybraca;
	ybracc=rootybfunc(bbracc, sln, qq[qp0], pp[qp0], 
			  ybracb, (3.0*ybracb+ymax)/4.0L, 1.0L/SMALL, (ymax-ybracb)/INTERVALS);
	if (fabsl(ybracc-ERROR)<SMALL) ymaxtmp=ymax; else ymaxtmp=ybracc;
	
	sol[qp0+1][2]=0.0L; sol[qp0+1][3]=0.0L;
	sol[qp0+1][4]=0.0L; sol[qp0+1][5]=0.0L;
	invquadfit(sol[qp0+1][1] - sol[qp0][1], ymintmp-sol[qp0][0], ymaxtmp-sol[qp0][0],
		   &sol[qp0][2], &sol[qp0][3]);
	invcubfit(sol[qp0+1][1] - sol[qp0][1], ymintmp-sol[qp0][0], ymaxtmp-sol[qp0][0],
		   &sol[qp0][4], &sol[qp0][5]);
    }
    sol[qpi][6]=sol[qpi-1][0]-sol[qpi][0];
    sol[qpi][7]=sol[qpi-1][1]-sol[qpi][1];
    
    
    for (qpi=qp0; qpi<=qp0+1; qpi+=1) {
	fprintf(fl1, "%5d%24.19Lf%24.19Lf%24.19Lf%24.19Lf%24.19Lf%10.3Lg%10.3Lg%5d\n", 
		qpi, a0, sol[qpi][1], sol[qpi][0], sol[qpi][7], sol[qpi][6], 
		sol[qpi][4], sol[qpi][5], okbrac);
    }
    
    
    for (qpi=qp0+2; qpi<=qpx; qpi+=1) {
	do {
	    tmp=qqh[qpi]; qqh[qpi]=qq[qpi]; qq[qpi]=tmp;  /* if not [..,1,1,1,...], try 1's 1st */
	    tmp=pph[qpi]; pph[qpi]=pp[qpi]; pp[qpi]=tmp;
	    
	    selectfiterfuncs(qpi,sln);
	    if (((qpi+innercurve)/2)*2==qpi+innercurve)  /* if not only 1's, sol[qpi] on 2nd try */
		ybracb=(pp[qpi]<=pph[qpi]? sol[qpi-2][0] :sol[qpi][0]);
	    else ybracb=sol[qpi-1][0];
	    ybraca=rootybfunc(bbracc, sln, qq[qpi], pp[qpi], 
			      ymintmp, (3.0L*ybracb+ymintmp)/4.0L, ybracb, (ybracb-ymintmp)/INTERVALS);
	    if (fabsl(ybraca-ERROR)<SMALL) ybraca=ymintmp; else ymintmp=(ybraca+=TINY);
	    ybracc=rootybfunc(bbracc, sln, qq[qpi], pp[qpi], 
			      ybracb, (3.0*ybracb+ymaxtmp)/4.0L, ymaxtmp, (ymaxtmp-ybracb)/INTERVALS);
	    if (fabsl(ybracc-ERROR)<SMALL) ybracc=ymaxtmp; else ymaxtmp=(ybracc-=TINY);
	    if((okbrac=mymnbrac(&ybraca, &ybracb, &ybracc, INTERVALS, rootbyfunc, 
				sln, qq[qpi], pp[qpi], bbraca, bbracc))>=1) {
		sol[qpi][1]=-minmax*brent(ybraca, ybracb, ybracc, rootbyfunc, 
					  sln, qq[qpi], pp[qpi], bbraca, bbracc, &sol[qpi][0]);
		if (((qpi+innercurve)/2)*2==qpi+innercurve) bbracc=sol[qpi][1]; 
		else bbraca=sol[qpi][1];
	    }	    
	} while (pp[qpi]<pph[qpi]);


	invquadfit( ((((qpi+innercurve)/2)*2==qpi+innercurve)?sol[qpi-2][1]:sol[qpi-1][1])-sol[qpi][1],
		    ymintmp-sol[qpi][0], ymaxtmp-sol[qpi][0],
		    &sol[qpi][2], &sol[qpi][3]);
	invcubfit( ((((qpi+innercurve)/2)*2==qpi+innercurve)?sol[qpi-2][1]:sol[qpi-1][1])-sol[qpi][1],
		   ymintmp-sol[qpi][0], ymaxtmp-sol[qpi][0],
		   &sol[qpi][4], &sol[qpi][5]);
	sol[qpi][6]=sol[qpi-1][0]-sol[qpi][0];
	sol[qpi][7]=sol[qpi-1][1]-sol[qpi][1];
	
	fprintf(fl1, "%5d%24.19Lf%24.19Lf%24.19Lf%24.19Lf%24.19Lf%10.3Lg%10.3Lg%5d\n", 
		qpi, a0, sol[qpi][1], sol[qpi][0], sol[qpi][7], sol[qpi][6], 
		sol[qpi][4], sol[qpi][5], okbrac);
	
	if (qpi==qpx-13 || qpi==qpx-12 || qpi==qpx-1 || qpi==qpx)
	    fprintf(fl3, "%5d%24.19Lf%24.19Lf%24.19Lf\n", 
		    qpi, a0, sol[qpi][1], sol[qpi][0]);

	fflush(fl1);
	fflush(fl3);
    }
    fprintf(fl1, "%5d%24.19Lf%24.19Lf%24.19Lf\n", 
		10*qpx-1, a0, (sol[qpx][1]+sol[qpx-1][1])/2.0L, (sol[qpx][0]+sol[qpx-1][0])/2.0L);
    if (qpx-qp0>12) {
	fprintf(fl1, "%5d%24.19Lf%24.19Lf%24.19Lf\n", 
		10*qpx, a0, sol[qpx-1][1]+(sol[qpx][1]-sol[qpx-1][1])*(sol[qpx-1][1]-sol[qpx-13][1])/
		((sol[qpx-12][1]-sol[qpx-13][1])-(sol[qpx][1]-sol[qpx-1][1])), 
		sol[qpx-1][0]+(sol[qpx][0]-sol[qpx-1][0])*(sol[qpx-1][0]-sol[qpx-13][0])/
		((sol[qpx-12][0]-sol[qpx-13][0])-(sol[qpx][0]-sol[qpx-1][0])));
	fprintf(fl2, "%24.19Lf%24.19Lf%24.19Lf\n", 
		a0, sol[qpx-1][1]+(sol[qpx][1]-sol[qpx-1][1])*(sol[qpx-1][1]-sol[qpx-13][1])/
		((sol[qpx-12][1]-sol[qpx-13][1])-(sol[qpx][1]-sol[qpx-1][1])), 
		sol[qpx-1][0]+(sol[qpx][0]-sol[qpx-1][0])*(sol[qpx-1][0]-sol[qpx-13][0])/
		((sol[qpx-12][0]-sol[qpx-13][0])-(sol[qpx][0]-sol[qpx-1][0])));
	fprintf(fl3, "%5s%24.19Lf%24.19Lf%24.19Lf\n", 
		"", a0, sol[qpx-1][1]+(sol[qpx][1]-sol[qpx-1][1])*(sol[qpx-1][1]-sol[qpx-13][1])/
		((sol[qpx-12][1]-sol[qpx-13][1])-(sol[qpx][1]-sol[qpx-1][1])),
		sol[qpx-1][0]+(sol[qpx][0]-sol[qpx-1][0])*(sol[qpx-1][0]-sol[qpx-13][0])/
		((sol[qpx-12][0]-sol[qpx-13][0])-(sol[qpx][0]-sol[qpx-1][0])));
    }
    fprintf(fl1, "\n");
    fprintf(fl3, "\n");
    fflush(fl2);
    }
    
}


void invquadfit(long double bdiff, long double ymindiff, long double ymaxdiff,
		long double *coeff1, long double *coeff2) 
{
    *coeff2=bdiff/(ymindiff-ymaxdiff)*(1.0L/ymindiff-1.0L/ymaxdiff);
    *coeff1=-(*coeff2)*ymindiff+bdiff/ymindiff;
}

long double quadfit(long double yyy, long double quadpars[6])
{
    long double tmp=(yyy-quadpars[0]);
    return (quadpars[1]+quadpars[2]*tmp+quadpars[3]*tmp*tmp);
}

void invcubfit(long double bdiff, long double ymindiff, long double ymaxdiff,
		long double *coeff1, long double *coeff2) 
{
    *coeff2=bdiff/(ymindiff-ymaxdiff)*(1.0L/ymindiff/ymindiff-1.0L/ymaxdiff/ymaxdiff);
    *coeff1=-(*coeff2)*ymindiff+bdiff/ymindiff/ymindiff;
}

long double cubfit(long double yyy, long double quadpars[6])
{
    long double tmp=(yyy-quadpars[0]);
    return (quadpars[1]+quadpars[4]*tmp*tmp+quadpars[5]*tmp*tmp*tmp);
}

long double qcfit(long double yyy, long double quadpars[6])
{
    long double tmp=(yyy-quadpars[0]), tmpq, tmpc, tmpu, tmpd;
    tmpq=(quadpars[1]+quadpars[2]*tmp+quadpars[3]*tmp*tmp);
    tmpc=(quadpars[1]+quadpars[4]*tmp*tmp+quadpars[5]*tmp*tmp*tmp);
    tmpu=((tmp=(tmpq>tmpc? tmpq : tmpc))>quadpars[1]? tmp : quadpars[1]);
    tmpd=((tmp=(tmpq<tmpc? tmpq : tmpc))>quadpars[1]? tmp : quadpars[1]);
    return (tmpu+tmpd)/2.0L;
}


long double rootybfunc(long double bbb, int sss, unsigned long int qqq, unsigned long int ppp,
		       long double y0tmp, long double yyy, long double yxxtmp, long double ybracsize0)
{
    int brakno=2, ok=0;
    long double bracl=yyy, bracu=yyy+ybracsize0;

    if ((*fiterfunc)(bbb, y0tmp, sss, qqq, ppp)*(*fiterfunc)(bbb, yxxtmp, sss, qqq, ppp)>= 0.0)
	ok=zbrac2(*fiterfunc, bbb, sss, qqq, ppp, y0tmp, yxxtmp, &bracl, &bracu);
    else { bracl=y0tmp; bracu=yxxtmp; ok=1;}
    if (ok>0) return zbrent2(*fiterfunc, bbb, sss, qqq, ppp, bracl, bracu);
    else return ERROR;
}

long double rootbyfunc(long double y, int sss, unsigned long int qqq, unsigned long int ppp,
		       long double b0tmp, long double bxxtmp)
{
    int brakno=2;

    if ((*fiterfunc)(b0tmp, y, sss, qqq, ppp)*(*fiterfunc)(bxxtmp, y, sss, qqq, ppp)>= 0.0)
	zbrak(*fiterfunc, y, sss, qqq, ppp, b0tmp, bxxtmp, bres, &brakl[0], &braku[0], &brakno);
    else { brakno=1; brakl[1]=b0tmp; braku[1]=bxxtmp; }
    if (brakno>0) return -minmax*zbrent(*fiterfunc, y, sss, qqq, ppp, brakl[brakno], braku[brakno]);
    else return 5.0L+fabsl(sqrtl(1.0L-omega[qpi]/a0)-fabsl(y)); 
/*    else return fabsl(y-10.0L*sqrtl(1.0L-omega/a0)); */
}

/* find x0(y0) so that point lies on symmetry line */
inline long double findx0(long double aa0, long double yy0, int symmln)
{
  long double tmp;

  switch (symmln) {
    case 1: {tmp=0.0L; break;}
    case 2: {tmp=0.5L; break;}
    case 3: {tmp=aa0/2.0L*(1.0L-yy0*yy0); break;}
    case 4: {tmp=aa0/2.0L*(1.0L-yy0*yy0)+0.5L; break;}
    default: {tmp=0.0L; printf("Uhoh - sln...\n");}
  }

  /* x modulo 1
  while (tmp> 0.5L-SMALL) tmp=tmp-1.0L;
  while (tmp< -0.5-SMALL) tmp=tmp+1.0L; */

  return tmp;
}

long double fiter2funcoe0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==1) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y; 
  for (i=1;i<=ppp/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-0.5L,xxtmp);}
  else {smoothen(xxtmp,xxtmp);}
  return xxtmp;
}

long double fiter2funcoe1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, j, jj, qf, pf;
  register long double xxtmp, yytmp;

  if (sss==3) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=ppp/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L,xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp),xxtmp);}
  return xxtmp;
}

long double fiter2funcoo0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==1) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp+1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L,xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp),xxtmp);}
  return xxtmp;
}

long double fiter2funcoo1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  if (sss==3) qf=(qqq-1)/2; else qf=(qqq+1)/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp-1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp-0.5L,xxtmp);}
  else {smoothen(xxtmp, xxtmp);}
  return xxtmp;
}

long double fiter2funceo0(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  qf=qqq/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp+1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==1) {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp), xxtmp);}
  else {smoothen(xxtmp-a0/2.0L*(1.0L-yytmp*yytmp)-0.5L, xxtmp);}
  return xxtmp;
}

long double fiter2funceo1(long double b, long double y, 
			  int sss, unsigned long int qqq, unsigned long int ppp) 
{
  register unsigned long int i, qf;
  register long double xxtmp, yytmp;

  qf=qqq/2;
  xxtmp=findx0(a0, y, sss); yytmp=y;
  for (i=1;i<=(ppp-1)/2;i++) {
    ntmapstep(xxtmp, yytmp, a0, b); 
    if (xxtmp>1.0L) if (qf>0) {xxtmp-=1.0L; qf--;}
  }
  xxtmp-=((long double)(qf));
  if (sss==3) {smoothen(xxtmp,xxtmp);}
  else {smoothen(xxtmp-0.5L, xxtmp);}
  return xxtmp;
}

void selectfiterfuncs(int sqpi, int ssln) 
{
      /* function selection for all possible q/p-combos */
      if (2*(pp[sqpi]/2)==pp[sqpi]) {            /* p even (-> q odd) */
	  if (ssln<=2) fiterfunc=&fiter2funcoe0;    /* I_0^i -> I_0^(1-i) after p/2, (q+-1)/2 */
	  else fiterfunc=&fiter2funcoe1;            /* I_1^i -> I_1^(i-1) after p/2, (q+-1)/2 */
      }
      else {                               /* p odd */
	  if (ssln<=2) {
	      if (2*(qq[sqpi]/2)==qq[sqpi])      /* q even */
		  fiterfunc=&fiter2funceo0;        /* I_0^i -> I_1^i after (p+1)/2, q/2 */
	      else fiterfunc=&fiter2funcoo0;       /* I_0^i -> I_1^(1-i) after (p+1)/2, (q+-1)/2 */
	  }
	  else {
	      if (2*(qq[sqpi]/2)==qq[sqpi])      /* q even */
		  fiterfunc=&fiter2funceo1;        /* I_0^1 -> I_1^1 after (p+1)/2, q/2 */
	      else fiterfunc=&fiter2funcoo1;       /* I_0^1 -> I_1^0 after (p+1)/2, (q+1)/2 */
	  }
      }
}

void mkstr(char *cnummer, unsigned long int nummer)
{
    unsigned long int tmp=nummer;
    char tmpstr[160], nostr[160];

    strcpy(tmpstr,"");
    while (tmp>0) {
	switch (tmp-10*(tmp/10)) {
	    case 1: {strcpy(nostr,"1"); break;}
	    case 2: {strcpy(nostr,"2"); break;}
	    case 3: {strcpy(nostr,"3"); break;}
	    case 4: {strcpy(nostr,"4"); break;}
	    case 5: {strcpy(nostr,"5"); break;}
	    case 6: {strcpy(nostr,"6"); break;}
	    case 7: {strcpy(nostr,"7"); break;}
	    case 8: {strcpy(nostr,"8"); break;}
	    case 9: {strcpy(nostr,"9"); break;}
	    case 0: {strcpy(nostr,"0"); break;}
	}
	strcat(nostr,tmpstr);
	strcpy(tmpstr,nostr);
      	tmp/=10;
    }
    strcpy(cnummer,tmpstr);
}

void zbrak(long double (*fx)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
	   long double x1, long double x2, int nn, 
	   long double xb1[], long double xb2[], int *nb)
{
  int nbb, i;
  long double x, fp, fc, dx;
  
  nbb=0;
  dx=(x2-x1)/nn;
  fp=(*fx)(x=x1, par0, par1, par2, par3);
  for (i=1;i<=nn;i++) {
    fc=(*fx)(x += dx, par0, par1, par2, par3);
    if (fc*fp <= 0.0) {
      xb1[++nbb]=x-dx;
      xb2[nbb]=x;
      if (*nb==nbb) return;
    }
    fp=fc;
  }
  *nb = nbb;
}

int zbrac2(long double (*func)(long double, long double, int, unsigned long int, unsigned long int), 
	   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
	   long double minbrac, long double maxbrac, long double *x1, long double *x2) 
{
    int j, NTRY=50;
    long double f1, f2, FACTOR=1.6, TOOBIG=0.1L/LDBL_EPSILON;

    if (*x1 == *x2) (*x2 += 10.0L*LDBL_EPSILON);
    if (*x1 > *x2) { f1=*x1; *x2=*x1; *x1=f1; }
    f1=(*func)(par0, *x1, par1, par2, par3);
    f2=(*func)(par0, *x2, par1, par2, par3);
    for (j=1;j<=NTRY;j++) {
	if (f1*f2 < 0.0) return 1;
	if (fabsl(f1) < fabsl(f2) && *x1+FACTOR*(*x1 - *x2)>minbrac)
	    f1=(*func)(par0, *x1 += FACTOR*(*x1 - *x2), par1, par2, par3);
	else
	    f2=(*func)(par0, *x2 += FACTOR*(*x2 - *x1), par1, par2, par3);
	if (fabsl(*x2)>TOOBIG || (*x2>maxbrac)) break;
    }
    return 0;
}



long double zbrent(long double (*func)(long double, long double, 
				       int, unsigned long int, unsigned long int), 
		   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
		   long double x1, long double x2)
{
    long double EPS=LDBL_EPSILON, ztol=LDBL_EPSILON;
    int ITMAX=100;

    int i;
    long double a=x1,b=x2,c=x2,d,e,min1,min2;
    long double fa, fb, fc,p,q,r,s,tol1,xm;

    fa = (*func)(a, par0, par1, par2, par3); fb = (*func)(b, par0, par1, par2, par3);
    if ((fa > 0.0L && fb > 0.0L) || (fa < 0.0L && fb < 0.0L)) return ERROR;
    fc=fb;
    for (i=1;i<=ITMAX;i++) {
	if ((fb > 0.0L && fc > 0.0L) || (fb < 0.0L && fc < 0.0L)) {
	    c=a;
	    fc=fa;
	    e=d=b-a;
	}
	if (fabsl(fc) < fabsl(fb)) {
	    a=b;
	    b=c;
	    c=a;
	    fa=fb;
	    fb=fc;
	    fc=fa;
	}
	tol1=2.0L*EPS*fabsl(b)+0.5L*ztol;
	xm=0.5L*(c-b);
	if (fabsl(xm) <= tol1 || fb == 0.0L) return b;
	if (fabsl(e) >= tol1 && fabsl(fa) > fabsl(fb)) {
	    s=fb/fa;
	    if (a == c) {
		p=2.0L*xm*s;
		q=1.0L-s;
	    } else {
		q=fa/fc;
		r=fb/fc;
		p=s*(2.0L*xm*q*(q-r)-(b-a)*(r-1.0L));
		q=(q-1.0L)*(r-1.0L)*(s-1.0L);
	    }
	    if (p > 0.0L) q = -q;
	    p=fabsl(p);
	    min1=3.0L*xm*q-fabsl(tol1*q);
	    min2=fabsl(e*q);
	    if (2.0L*p < (min1 <min2 ? min1 : min2)) {
		e=d;
		d=p/q;
	    } else {
		d=xm;
		e=d;
	    }
	} else {
	    d=xm;
	    e=d;
	}
	a=b;
	fa=fb;
	if (fabsl(d) > tol1)
	    b += d;
	else
	    b += (xm > 0.0L ? fabsl(tol1) : -fabsl(tol1));
	fb=(*func)(b, par0, par1, par2, par3);
    }
    return ERROR;
}

long double zbrent2(long double (*func)(long double, long double, 
				       int, unsigned long int, unsigned long int), 
		   long double par0, int par1, unsigned long int par2, unsigned long int par3, 
		   long double x1, long double x2)
{
    long double EPS=LDBL_EPSILON, ztol=LDBL_EPSILON;
    int ITMAX=100;

    int i;
    long double a=x1,b=x2,c=x2,d,e,min1,min2;
    long double fa, fb, fc,p,q,r,s,tol1,xm;

    fa = (*func)(par0, a, par1, par2, par3); fb = (*func)(par0, b, par1, par2, par3);
    if ((fa > 0.0L && fb > 0.0L) || (fa < 0.0L && fb < 0.0L)) return ERROR;
    fc=fb;
    for (i=1;i<=ITMAX;i++) {
	if ((fb > 0.0L && fc > 0.0L) || (fb < 0.0L && fc < 0.0L)) {
	    c=a;
	    fc=fa;
	    e=d=b-a;
	}
	if (fabsl(fc) < fabsl(fb)) {
	    a=b;
	    b=c;
	    c=a;
	    fa=fb;
	    fb=fc;
	    fc=fa;
	}
	tol1=2.0L*EPS*fabsl(b)+0.5L*ztol;
	xm=0.5L*(c-b);
	if (fabsl(xm) <= tol1 || fb == 0.0L) return b;
	if (fabsl(e) >= tol1 && fabsl(fa) > fabsl(fb)) {
	    s=fb/fa;
	    if (a == c) {
		p=2.0L*xm*s;
		q=1.0L-s;
	    } else {
		q=fa/fc;
		r=fb/fc;
		p=s*(2.0L*xm*q*(q-r)-(b-a)*(r-1.0L));
		q=(q-1.0L)*(r-1.0L)*(s-1.0L);
	    }
	    if (p > 0.0L) q = -q;
	    p=fabsl(p);
	    min1=3.0L*xm*q-fabsl(tol1*q);
	    min2=fabsl(e*q);
	    if (2.0L*p < (min1 <min2 ? min1 : min2)) {
		e=d;
		d=p/q;
	    } else {
		d=xm;
		e=d;
	    }
	} else {
	    d=xm;
	    e=d;
	}
	a=b;
	fa=fb;
	if (fabsl(d) > tol1)
	    b += d;
	else
	    b += (xm > 0.0L ? fabsl(tol1) : -fabsl(tol1));
	fb=(*func)(par0, b, par1, par2, par3);
    }
    return ERROR;
}


void mnbrac(long double *ax, long double *bx, long double *cx, 
	    long double (*func)(long double, int, unsigned long int, unsigned long int, long double, long double), 
		   int par1, unsigned long int par2, unsigned long int par3, long double par4, long double par5)
{
  long double GLIMIT=10.0L, EPS=LDBL_EPSILON, TINY=EPS;
  long double GOLD=1.6180339887498948482045868343656381177203091798058L;
  
  long double ulim, u, r, q, fu, dum;
  long double fa, fb, fc;

  fa = (*func)(*ax, par1, par2, par3, par4, par5); fb = (*func)(*bx, par1, par2, par3, par4, par5);
  if (fb > fa) {
    SHFT(dum, *ax, *bx, dum);
    SHFT(dum, fb, fa, dum);
  }
  *cx = (*bx) + GOLD*(*bx-*ax);
  fc = (*func)(*cx, par1, par2, par3, par4, par5);
  while (fb > fc) {
    r = (*bx - *ax)*(fb-fc);
    q = (*bx - *cx)*(fb-fa);
    u = (*bx) - 
      ((*bx - *cx)*q - (*bx - *ax)*r)/(2.0*SIGN(FMAX(fabsl(q-r), TINY),q-r));
    ulim = (*bx) + GLIMIT*(*cx-*bx);
    if ((*bx-u)*(u-*cx) > 0.0) {
      fu = (*func)(u, par1, par2, par3, par4, par5);
      if (fu < fc) {
	*ax = (*bx);
	*bx = u;
	fa = fb;
	return;
      } else if (fu > fb) {
	*cx = u;
	fc = fu;
	return;
      }
      u = (*cx) + GOLD*(*cx-*bx);
      fu = (*func)(u, par1, par2, par3, par4, par5);
    } else if ((*cx-u)*(u-ulim) > 0.0) {
      fu = (*func)(u, par1, par2, par3, par4, par5);
      if (fu < fc) {
	SHFT(*bx, *cx, u, *cx+GOLD*(*cx-*bx));
	SHFT(fb, fc, fu, (*func)(u, par1, par2, par3, par4, par5));
      }
    } else if ((u-ulim)*(ulim-*cx) >= 0.0) {
      u = ulim;
      fu = (*func)(u, par1, par2, par3, par4, par5);
    } else {
      u = (*cx) + GOLD*(*cx-*bx);
      fu = (*func)(u, par1, par2, par3, par4, par5);
    }
    SHFT(*ax, *bx, *cx, u);
    SHFT(fa, fb, fc, fu);
  }
}

int mymnbrac(long double *ax, long double *bx, long double *cx, long double distf,
	    long double (*func)(long double, int, unsigned long int, unsigned long int, long double, long double), 
		   int par1, unsigned long int par2, unsigned long int par3, long double par4, long double par5)
/* find minimum bracket between *ax and *cx, i.e. f(*bx)<f(*ax),
f(*bx)<f(*cx), where *ax and *cx must not be exceeded; 
return values: 1 - minimum found, 2 - boundary found, 0 - bracket size too small, 
-i - no convergence after i attempts, -1 - nothing worked */
{
  long double GLIMIT=5.0L, EPS=LDBL_EPSILON, TINY=EPS;
  long double GOLD=1.6180339887498948482045868343656381177203091798058L;
  int NTRY=10;
 
  long double ulim, u, r, q, fu, dum;
  long double fa, fb, fc, a0=*ax, b0=*bx, c0=*cx, distx;
  int i;

  if (fabsl(*ax-*cx)<5.0L*TINY) return 0;          /* bracket size too small */ 
  distx=(*cx-*ax)/distf; if (fabsl(distx)<5.0L*TINY) distx=5.0L*TINY;
  if ((*ax-*bx)*(*bx-*cx)<0) *bx=(*ax + *cx)/2.0L; b0=*bx; /* need *bx inside bracket */
    
  fa = (*func)(a0, par1, par2, par3, par4, par5); 
  fb = (*func)(b0, par1, par2, par3, par4, par5);
  fc = (*func)(c0, par1, par2, par3, par4, par5);
  if (fa>fb && fc>fb) {                            /* bracket already found */
      if (fa<=5.0L && fc <=5.0L) return 1;
      else  {
	  if (fa> 5.0L) {
	      *ax=(3.0L*a0+b0)/4.0L;
	      fa = (*func)(*ax, par1, par2, par3, par4, par5); 
	  }
	  if (fc> 5.0L) {
	      *cx=(3.0L*c0+b0)/4.0L;
	      fc = (*func)(*cx, par1, par2, par3, par4, par5); 
	  }
	  if (fa>fb && fc>fb) return ((fa<=5.0L && fc<=5.0L)? 1: 2);
	  else {
	      *ax=a0; *cx=c0;
	      return 2;
	  }
      }
  }
  
  *ax=b0; *bx=b0+distx/2.0L; *cx=b0+distx;         /* try search from center of bracket */
  fa = (*func)(*ax, par1, par2, par3, par4, par5); 
  fb = (*func)(*bx, par1, par2, par3, par4, par5);
  fc = (*func)(*cx, par1, par2, par3, par4, par5);
  if (fa>fb && fc>fb) return ((fa<=5.0L && fc<=5.0L)? 1: 2);     /* bracket inside start interval */

  *bx=*cx; fb=fc;
  if (fb > fa) {
    SHFT(dum, *ax, *bx, dum);
    SHFT(dum, fb, fa, dum);
    SHFT(dum, a0, c0, dum);
    distx=-distx;
  }
  
  *cx = (*bx) + GOLD*(*bx-*ax);
  fc = (*func)(*cx, par1, par2, par3, par4, par5);

  for (i=1;i<=NTRY;i++) {
      while (fb > fc && (c0-*cx)*(*cx-*bx)>0.0L) {
	  r = (*bx - *ax)*(fb-fc);
	  q = (*bx - *cx)*(fb-fa);
	  u = (*bx) - 
	      ((*bx - *cx)*q - (*bx - *ax)*r)/(2.0*SIGN(FMAX(fabsl(q-r), TINY),q-r));
	  ulim = (*bx) + GLIMIT*(*cx-*bx);               /* allowed limit for interpolation */
	  if ((c0-ulim)*(*cx-*bx)<0.0L) ulim=c0-distx;
	  if ((*bx-u)*(u-*cx) > 0.0) {                   /* parabolic u between *bx and *cx */
	      fu = (*func)(u, par1, par2, par3, par4, par5);
	      if (fu < fc) {
		  *ax = (*bx);
		  *bx = u;
		  fa = fb;
		  return ((fa<=5.0L && fc<=5.0L)? 1: 2);
	      } else if (fu > fb) {
		  *cx = u;
		  fc = fu;
		  return ((fa<=5.0L && fc<=5.0L)? 1: 2);
	      }
	      u = (*cx) + GOLD*(*cx-*bx);
	      fu = (*func)(u, par1, par2, par3, par4, par5);
	  } else if ((*cx-u)*(u-ulim) > 0.0) {           /* parabolic u between *cx and limit */
	      fu = (*func)(u, par1, par2, par3, par4, par5);
	      if (fu < fc) {
		  SHFT(*bx, *cx, u, *cx+GOLD*(*cx-*bx));
		  SHFT(fb, fc, fu, (*func)(u, par1, par2, par3, par4, par5));
	      }
	  } else if ((u-ulim)*(ulim-*cx) >= 0.0) {
	      u = ulim;
	      fu = (*func)(u, par1, par2, par3, par4, par5);
	  } else {
	      u = (*cx) + GOLD*(*cx-*bx);
	      fu = (*func)(u, par1, par2, par3, par4, par5);
	  }
	  SHFT(*ax, *bx, *cx, u);
	  SHFT(fa, fb, fc, fu);
      }
      if (fb < fc && (c0-*cx)*(*cx-*bx)>0.0L) return ((fa<=5.0L && fc<=5.0L)? 1: 2);
      if ((c0-*cx)*(*cx-*bx)<=0.0L) {
	  *cx=c0; 
	  fc = (*func)(*cx, par1, par2, par3, par4, par5);
	  if (fb < fc) return ((fa <=5.0L && fc<=5.0L)? 1: 2);
      }
      
      c0=b0; b0=(a0+c0)/2.0L;
      if (i==NTRY) {                                   /* last try */ 
	  *ax=a0; *bx=a0+distx/2.0L; *cx=a0+distx;
	  if ((*cx-c0)*(*cx-*bx)>0.0) {
	      *cx=c0; *bx=(*ax+*cx)/2.0L;
	  }
	  fa = (*func)(*ax, par1, par2, par3, par4, par5); 
	  fb = (*func)(*bx, par1, par2, par3, par4, par5);
	  fc = (*func)(*cx, par1, par2, par3, par4, par5);
	  if (fa>fb && fc>fb) return ((fa <=5.0L && fc<=5.0L)? 1: 2); /* bracket inside start interval */
      
	  *bx=*cx; fb=fc;
	  if (fb>fa) return -i;                        /* wrong bracket direction */
      }
      
      if (i<NTRY) {                                    /* try again, in ignored end of bracket */ 
	  *ax=b0; *bx=*ax+distx/2.0L; *cx=*ax+distx;
	  if ((*cx-c0)*(*cx-*bx)>0.0) {
	      i=NTRY;
	      *cx=c0; *bx=(*ax+*cx)/2.0L;
	  }
	  fa = (*func)(*ax, par1, par2, par3, par4, par5); 
	  fb = (*func)(*bx, par1, par2, par3, par4, par5);
	  fc = (*func)(*cx, par1, par2, par3, par4, par5);
	  if (fa>fb && fc>fb) return ((fa<=5.0L && fc<=5.0L)? 1: 2); /* bracket inside start interval */
	  *bx=*cx; fb=fc;
	  if (fb > fa) {
	      SHFT(dum, *ax, *bx, dum);
	      SHFT(dum, fb, fa, dum);
	      SHFT(dum, a0, c0, dum);
	      distx=-distx;
	  }
      }
  
      *cx = (*bx) + GOLD*(*bx-*ax);
      fc = (*func)(*cx, par1, par2, par3, par4, par5);
  }
  if (fb < fc && (c0-*cx)*(*cx-*bx)>0.0L) return ((fa<=5.0L && fc<=5.0L)? 1: 2);
  else return -1;
}

long double brent(long double ax, long double bx, long double cx, 
	    long double (*f)(long double, int, unsigned long int, unsigned long int, long double, long double), 
		  int par1, unsigned long int par2, unsigned long int par3, long double par4, long double par5,
		  long double *xmin)
{
  long double ITMAX=100, ZEPS=LDBL_EPSILON, tol=sqrtl(ZEPS);
  long double CGOLD=0.3819660112501051517954131656343618822796908201942L;
  int iter;
  long double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm,e=0.0L;

  a=(ax < cx ? ax : cx);
  b=(ax > cx ? ax : cx);
  x=w=v=bx;
  fw=fv=fx=(*f)(x, par1, par2, par3, par4, par5); 
  for (iter=1; iter<=ITMAX; iter++) {
    xm=0.5L*(a+b);
    tol2= 2.0L*(tol1=tol*fabsl(x)+ZEPS);
    if (fabsl(x-xm) <= (tol2-0.5L*(b-a))) {
      *xmin = x;
      return fx;
    }
    if (fabsl(e) > tol1) {
      r = (x-w)*(fx-fv);
      q = (x-v)*(fx-fw);
      p = (x-v)*q - (x-w)*r;
      q = 2.0L*(q-r);
      if (q > 0.0L) p = -p;
      q = fabsl(q);
      etemp=e;
      e=d;
      if (fabsl(p) >= fabsl(0.5L*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
	d = CGOLD*(e=(x >= xm ? a-x : b-x));
      else {
	d = p/q;
	u = x+d;
	if (u-a < tol2 || b-u < tol2)
	  d = SIGN(tol1,xm-x);
      }
    } else {
      d = CGOLD*(e=(x >= xm ? a-x : b-x));
    }
    u = (fabsl(d) >= tol1 ? x+d : x+SIGN(tol1,d));
    fu = (*f)(u, par1, par2, par3, par4, par5); 
    if (fu <= fx) {
      if (u >= x) a = x; else b = x;
      SHFT(v,w,x,u);
      SHFT(fv,fw,fx,fu);
    } else {
      if (u < x) a = u; else b = u;
      if (fu <= fw || w == x) {
	v=w;
	w=u;
	fv=fw;
	fw=fu;
      } else if (fu <= fv || v == x || v == w) {
	v = u;
	fv = fu;
      }
    }
  }
  *xmin = x;
  return ERROR;
}

int qsortcompare(const void *a, const void *b)
{
  return (((long double *)a)[0]<((long double *)b)[0]? 
	  1 : (((long double *)a)[0]>((long double *)b)[0] ? -1 : 0));
}
